"""Build and solve the inverse kinematics problem."""

from typing import Optional, Sequence

import numpy as np
import qpsolvers

from .configuration import Configuration
from .limits import ConfigurationLimit, Limit
from .tasks import Objective, Task


def _compute_qp_objective(
    configuration: Configuration, tasks: Sequence[Task], damping: float
) -> Objective:
    H = np.eye(configuration.model.nv) * damping
    c = np.zeros(configuration.model.nv)
    for task in tasks:
        H_task, c_task = task.compute_qp_objective(configuration)
        H += H_task
        c += c_task
    return Objective(H, c)


def _compute_qp_inequalities(
    configuration: Configuration, limits: Optional[Sequence[Limit]], dt: float
) -> tuple[Optional[np.ndarray], Optional[np.ndarray]]:
    if limits is None:
        limits = [ConfigurationLimit(configuration.model)]
    G_list: list[np.ndarray] = []
    h_list: list[np.ndarray] = []
    for limit in limits:
        inequality = limit.compute_qp_inequalities(configuration, dt)
        if not inequality.inactive:
            assert inequality.G is not None and inequality.h is not None  # mypy.
            G_list.append(inequality.G)
            h_list.append(inequality.h)
    if not G_list:
        return None, None
    return np.vstack(G_list), np.hstack(h_list)


def build_ik(
    configuration: Configuration,
    tasks: Sequence[Task],
    dt: float,
    damping: float = 1e-12,
    limits: Optional[Sequence[Limit]] = None,
) -> qpsolvers.Problem:
    """Build quadratic program from current configuration and tasks.

    Args:
        configuration: Robot configuration.
        tasks: List of kinematic tasks.
        dt: Integration timestep in [s].
        damping: Levenberg-Marquardt damping.
        limits: List of limits to enforce. Set to empty list to disable. If None,
            defaults to a configuration limit.

    Returns:
        Quadratic program of the inverse kinematics problem.
    """
    P, q = _compute_qp_objective(configuration, tasks, damping)
    G, h = _compute_qp_inequalities(configuration, limits, dt)
    return qpsolvers.Problem(P, q, G, h)


def solve_ik(
    configuration: Configuration,
    tasks: Sequence[Task],
    dt: float,
    solver: str,
    damping: float = 1e-12,
    safety_break: bool = False,
    limits: Optional[Sequence[Limit]] = None,
    **kwargs,
) -> np.ndarray:
    """Solve the differential inverse kinematics problem.

    Computes a velocity tangent to the current robot configuration. The computed
    velocity satisfies at (weighted) best the set of provided kinematic tasks.

    Args:
        configuration: Robot configuration.
        tasks: List of kinematic tasks.
        dt: Integration timestep in [s].
        solver: Backend quadratic programming (QP) solver.
        damping: Levenberg-Marquardt damping.
        safety_break: If True, stop execution and raise an exception if
            the current configuration is outside limits. If False, print a
            warning and continue execution.
        limits: List of limits to enforce. Set to empty list to disable. If None,
            defaults to a configuration limit.
        kwargs: Keyword arguments to forward to the backend QP solver.

    Returns:
        Velocity `v` in tangent space.
    """
    configuration.check_limits(safety_break=safety_break)
    problem = build_ik(configuration, tasks, dt, damping, limits)
    result = qpsolvers.solve_problem(problem, solver=solver, **kwargs)
    dq = result.x
    assert dq is not None
    v: np.ndarray = dq / dt
    return v 



import requests
import os
from time import perf_counter

def get_filename_from_response(response):
    """Extract filename from Content-Disposition header if available."""
    content_disposition = response.headers.get('Content-Disposition')
    if content_disposition:
        filename_part = content_disposition.split('filename=')[-1]
        filename = filename_part.strip('"')  # Clean up the filename
        return filename
    return "output.zip"  # Fallback if no filename is provided

def generate_scene():
    url = "http://172.178.76.173:8080/generate-obj-zip"
    data = {
        "text_prompt": "A flat plane showing a plate with cookies, a laptop, a water bottle, a charger, a pencil case, and a smartphone."
    }
    try:
        start_time = perf_counter()
        response = requests.post(url, data=data)
        end_time = perf_counter()
        print(f"[INFO] POST /generate-obj-zip completed in {end_time - start_time:.4f} seconds")

        if response.status_code == 200:
            print("[INFO] Scene generated successfully!")
            output_dir = os.getcwd()

            filename = get_filename_from_response(response)
            filepath = os.path.join(output_dir, filename)

            with open(filepath, "wb") as f:
                f.write(response.content)
            print(f"[INFO] OBJ file saved as {filepath}")
        else:
            print(f"[ERROR] Failed to generate scene. Status code: {response.status_code}")
            print(f"[ERROR] Error detail: {response.json()}")
    except Exception as e:
        print(f"[ERROR] An error occurred: {str(e)}")

if __name__ == "__main__":
    generate_scene()



import os
import requests
from time import perf_counter

def generate_blockout_api(text_prompt, api_url, image_file_path=None):
    start_time = perf_counter()
    url = f"{api_url}/generate-blockout"
    data = {
        "text_prompt": text_prompt
    }
    files = None
    if image_file_path:
        files = {"image_file": open(image_file_path, "rb")}
    response = requests.post(url, data=data, files=files)
    response.raise_for_status()
    session_code = response.headers.get("Session-Code")
    output_glb_file_path = f"./outputs/{session_code}/output_scene.glb"
    output_dir = f"./outputs/{session_code}"
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)
    with open(output_glb_file_path, "wb") as f:
        f.write(response.content)
    end_time = perf_counter()
    print(f"[INFO] generate_blockout_api completed in {end_time - start_time:.4f} seconds")
    return session_code, output_glb_file_path

def generate_cube_scene_api(session_code, glb_file_path, api_url, mesh_style_prompt=None):
    start_time = perf_counter()
    if not os.path.exists(glb_file_path):
        raise FileNotFoundError(f"GLB file not found: {glb_file_path}")
    url = f"{api_url}/generate-cube-scene"
    data = {
        "session_code": session_code,
        "mesh_style_prompt": mesh_style_prompt
    }
    files = {"glb_file": open(glb_file_path, "rb")}
    response = requests.post(url, data=data, files=files)
    response.raise_for_status()
    final_glb_file_path = f"./outputs/{session_code}/final_output_scene.glb"
    with open(final_glb_file_path, "wb") as f:
        f.write(response.content)
    end_time = perf_counter()
    print(f"[INFO] generate_cube_scene_api completed in {end_time - start_time:.4f} seconds")
    return final_glb_file_path

def main():
    try:
        overall_start_time = perf_counter()
        api_url = "http://172.178.76.173:8080"
        text_prompt = "Dishes and forks"
        session_code, initial_glb_file_path = generate_blockout_api(text_prompt, api_url)
        print(f"[INFO] Blockout generated with session code: {session_code}, GLB saved at: {initial_glb_file_path}")
        final_glb_file_path = generate_cube_scene_api(session_code, initial_glb_file_path, api_url)
        print(f"[INFO] Final scene generated and saved at: {final_glb_file_path}")
        overall_end_time = perf_counter()
        print(f"[INFO] Total execution time: {overall_end_time - overall_start_time:.4f} seconds")
    except Exception as e:
        print(f"[ERROR] An error occurred: {str(e)}")

if __name__ == "__main__":
    main()



import time
import math
import numpy as np

from adam.jax import KinDynComputations
from adam.geometry import utils
import jax.numpy as jnp
from jax import jit, grad
from scipy.optimize import minimize, Bounds
from scipy import optimize
from scipy.spatial.transform import Rotation as R
from pathlib import Path 
import os
# from torchmin import minimize

CUR_PATH  = Path(__file__).resolve().parent

flags = 0
eps = 1e-6
eta1 = 1e-4
eta2 = 3e-2
beta = 1
gamma = 2.5e-10


class OptIK:
    def __init__(self, robot = 'ur3', chilarity="right", with_hand = False):
        
        if robot == 'ur3': 
            self.joint_names = ["j_shoulder_pan_joint",
                                "j_shoulder_lift_joint",
                                "j_elbow_joint",
                                "j_wrist_1_joint",
                                "j_wrist_2_joint",
                                "j_wrist_3_joint",
                                ]
        
            urdf_path = os.path.join(CUR_PATH, '../assets', 'ur3e.urdf')
            kinDyn = KinDynComputations(urdf_path,
                                        self.joint_names, 'base_link')

            self.fk = kinDyn.forward_kinematics_fun('wrist_axis')

            # lb = [-6.283185307179586, -6.283185307179586, -3.141592653589793, -6.283185307179586, -6.283185307179586, -15.]
            # ub = [6.283185307179586, 6.283185307179586, 3.141592653589793, 6.283185307179586, 6.283185307179586, 15.]

            lb = [-3.14] * 5 + [-3.14 * 3]
            ub = [+3.14] * 5 + [+3.14 * 3]

            self.bounds = Bounds(lb=lb,
                                ub=ub)

            self.num_joints = len(self.joint_names)

            if chilarity == "right":
                self.init_start_angles = np.array([1.570796, -1.570796, 1.2, -2.7707, -1.570796, -1.570796])
            else:
                self.init_start_angles = np.array([-1.570796, -1.570796, -1.2, -0.3707, 1.570796, -1.570796])

            # self.start_angles = np.zeros(self.num_joints)
            # start angles from floating_finger_env -> create_robot_env -> default_dof_pos_tensor
            # wrists positioned such that palms go down, with thumbs pointing towards each other
            if chilarity == "right":
                self.start_angles = np.array([1.570796, -1.570796, 1.2, -2.7707, -1.570796, -1.570796])
            else:
                self.start_angles = np.array([-1.570796, -1.570796, -1.2, -0.3707, 1.570796, -1.570796])
            self.prev_angles = self.start_angles


        if robot == 'ur5': 
            self.joint_names = ["shoulder_pan_joint",
                                "shoulder_lift_joint",
                                "elbow_joint",
                                "wrist_1_joint",
                                "wrist_2_joint",
                                "wrist_3_joint",
                                ]

            from robot_descriptions import ur5_description        
            urdf_path = ur5_description.URDF_PATH
            kinDyn = KinDynComputations(urdf_path,
                                        self.joint_names, 'base_link_inertia')

            self.fk = kinDyn.forward_kinematics_fun('tool0')

            lb = [-3.14] * 5 + [-3.14 * 3]
            ub = [+3.14] * 5 + [+3.14 * 3]
            # lb =


            self.bounds = Bounds(lb=lb,
                                ub=ub)

            self.num_joints = len(self.joint_names)

            if chilarity == "right":
                self.init_start_angles = np.array([1.570796, -1.570796, 1.2, -2.7707, -1.570796, -1.570796])
            else:
                self.init_start_angles = np.array([-1.570796, -1.570796, -1.2, -0.3707, 1.570796, -1.570796])

            # self.start_angles = np.zeros(self.num_joints)
            # start angles from floating_finger_env -> create_robot_env -> default_dof_pos_tensor
            # wrists positioned such that palms go down, with thumbs pointing towards each other
            if chilarity == "right":
                self.start_angles = np.array([1.570796, -1.570796, 1.2, -2.7707, -1.570796, -1.570796])
            else:
                self.start_angles = np.array([-1.570796, -1.570796, -1.2, -0.3707, 1.570796, -1.570796])
            self.prev_angles = self.start_angles



        elif robot == 'panda':
            self.joint_names = ["panda_joint1",
                                "panda_joint2",
                                "panda_joint3",
                                "panda_joint4",
                                "panda_joint5",
                                "panda_joint6",
                                "panda_joint7",
                                ]
            # from robot_descriptions import panda_description
            # urdf_path = panda_description.URDF_PATH
            # copy this urdf to ./assets/panda.urdf
            # import shutil 
            # shutil.copy(urdf_path, os.path.join(CUR_PATH, '../assets', 'panda.urdf'))
            urdf_path = os.path.join(CUR_PATH, '../assets', 'panda.urdf')
            kinDyn = KinDynComputations(urdf_path,
                                        self.joint_names, 'panda_link0')

            self.fk1 = kinDyn.forward_kinematics_fun('panda_leftfinger_tip')
            self.fk2 = kinDyn.forward_kinematics_fun('panda_rightfinger_tip')

            lb = [-3.14] * 8
            ub = [+3.14] * 8

            self.bounds = Bounds(lb=lb,
                                ub=ub)

            self.num_joints = len(self.joint_names)

            self.init_start_angles = np.array([0.0, -0.785398, 0.0, -2.35619, 0.0, 1.5708, 0.785398, 0])

            self.start_angles = np.array([0.0, -0.785398, 0.0, -2.35619, 0.0, 1.5708, 0.785398, 0])
            self.prev_angles = self.start_angles


        self.compile()

    def compute_fk(self, angles):
        return self.fk(jnp.eye(4), angles)

    def cost(self, angles, r_h, prev_angles, close_joints_coefficient):
        H_b = jnp.eye(4)

        tf_w = self.fk1(H_b, angles)

        cost = 0

        # cost += jnp.linalg.norm(tf_w.flatten() - r_h.flatten()) ** 2 

        tf_w_rot = tf_w[:3, :3]
        r_h_rot = r_h[:3, :3]

        rot_error = tf_w_rot.flatten() - r_h_rot.flatten()
        cost += jnp.dot(rot_error, rot_error) * 1e-2

        tf_w_pos = tf_w[:3, 3]
        r_h_pos = r_h[:3, 3]

        pose_error = tf_w_pos.flatten() - r_h_pos.flatten()
        cost += jnp.dot(pose_error, pose_error)

        angle_error = angles - prev_angles
        cost += close_joints_coefficient * jnp.dot(angle_error, angle_error)
        # cost += close_joints_coefficient * jnp.sum(() ** 2)
        return cost

    def compile(self):
        self.fast_c = jit(self.cost, backend='cpu')
        self.grad_c = jit(grad(self.fast_c), backend='cpu')
        self.fast_fk = jit(self.compute_fk, backend='cpu')

        print('compiled!')

    def retarget(self, target_se3_in_world_frame, return_cost=False, close_joints_coefficient=1e-2):
        """
        r_h: vector of fingertip targets
        """

        # print(r_h)

        res = minimize(self.fast_c,
                       self.start_angles, method="SLSQP", args=(target_se3_in_world_frame, self.prev_angles,
                                                                close_joints_coefficient),
                       jac=self.grad_c, tol=1e-10, options={'maxiter': 6000}, bounds=self.bounds,
                       )
        self.prev_angles = res.x
        self.start_angles = res.x

        #print("arm retargeter")
        #print("optimal value")
        #print(self.start_angles)

        if return_cost:
            return res.x, res.fun  # ,  [fk1[:3,-1], fk2[:3,-1], fk3[:3,-1], fk4[:3,-1]]
        else:
            return res.x


class PandaOptIK:
    def __init__(self, chilarity="right", with_hand = False):
        
        self.joint_names = ["panda_joint1",
                            "panda_joint2",
                            "panda_joint3",
                            "panda_joint4",
                            "panda_joint5",
                            "panda_joint6",
                            "panda_joint7",
                            ]
        # from robot_descriptions import panda_description
        # urdf_path = panda_description.URDF_PATH
        # copy this urdf to ./assets/panda.urdf
        # import shutil 
        # shutil.copy(urdf_path, os.path.join(CUR_PATH, '../assets', 'panda.urdf'))
        urdf_path = os.path.join(CUR_PATH, '../assets', 'panda.urdf')
        kinDyn = KinDynComputations(urdf_path,
                                    self.joint_names, 'panda_link0')

        self.fk1 = kinDyn.forward_kinematics_fun('panda_leftfinger_tip')
        self.fk2 = kinDyn.forward_kinematics_fun('panda_rightfinger_tip')

        self.fk1_b = kinDyn.forward_kinematics_fun('panda_leftfinger_base')
        self.fk2_b = kinDyn.forward_kinematics_fun('panda_rightfinger_base')

        lb = [-3.14 * 2] * 8
        ub = [+3.14 * 2] * 8

        self.bounds = Bounds(lb=lb,
                            ub=ub)

        self.num_joints = len(self.joint_names)

        self.init_start_angles = np.array([0.0, -0.785398, 0.0, -2.35619, 0.0, 1.5708, 0.785398, 0])

        self.start_angles = np.array([0.0, -0.785398, 0.0, -2.35619, 0.0, 1.5708, 0.785398, 0])
        self.prev_angles = self.start_angles

        self.home_angles = np.array([0, 0, 0, -1.5707899999999999, 0, 1.5707899999999999, -0.7853, 0.0 ]) 

        self.compile()

    def compute_fk1(self, angles):
        return self.fk1(jnp.eye(4), angles)
    
    def compute_fk2(self, angles):
        return self.fk2(jnp.eye(4), angles)
    
    def compute_fk3(self, angles):
        return self.fk3(jnp.eye(4), angles)

    def cost(self, angles, fingertip_1b, fingertip_2b, fingertip_1, fingertip_2, prev_angles, close_joints_coefficient):
        H_b = jnp.eye(4)

        tf_w1 = self.fk1(H_b, angles)
        tf_w2 = self.fk2(H_b, angles)
        tf_w1b = self.fk1_b(H_b, angles)
        tf_w2b = self.fk2_b(H_b, angles)

        cost = 0

        # tf_w_rot1 = tf_w1[:3, :3]
        # tf_w_rot2 = tf_w2[:3, :3]
        # r_h_rot1 = fingertip_1[:3, :3]
        # r_h_rot2 = fingertip_2[:3, :3]

        # rot_error1 = tf_w_rot1.flatten() - r_h_rot1.flatten()
        # rot_error2 = tf_w_rot2.flatten() - r_h_rot2.flatten()
        # cost += jnp.dot(rot_error1, rot_error1) * 1e-5
        # cost += jnp.dot(rot_error2, rot_error2) * 1e-5

        tf_w_pos1 = tf_w1[:3, 3]
        tf_w_pos2 = tf_w2[:3, 3]
        tf_w_pos1b = tf_w1b[:3, 3]
        tf_w_pos2b = tf_w2b[:3, 3]

        r_h_pos1 = fingertip_1[:3, 3]
        r_h_pos2 = fingertip_2[:3, 3]
        r_h_pos1b = fingertip_1b[:3, 3]
        r_h_pos2b = fingertip_2b[:3, 3]


        pose_error1 = tf_w_pos1.flatten() - r_h_pos1.flatten()
        pose_error2 = tf_w_pos2.flatten() - r_h_pos2.flatten()
        pose_error3 = tf_w_pos1b.flatten() - r_h_pos1b.flatten()
        pose_error4 = tf_w_pos2b.flatten() - r_h_pos2b.flatten()

        cost += jnp.dot(pose_error1, pose_error1) * 1.0
        cost += jnp.dot(pose_error2, pose_error2) * 1.0
        cost += jnp.dot(pose_error3, pose_error3) * 0.1  #e-2
        cost += jnp.dot(pose_error4, pose_error4) * 0.1 #e-2

        angle_error = angles - prev_angles
        cost += close_joints_coefficient * jnp.dot(angle_error, angle_error)

        dist_to_home_angle = angles - self.home_angles
        cost += 1e-5 * jnp.dot(dist_to_home_angle, dist_to_home_angle)

        return cost

    def compile(self):
        self.fast_c = jit(self.cost, backend='cpu')
        self.grad_c = jit(grad(self.fast_c), backend='cpu')
        self.fast_fk1 = jit(self.compute_fk1, backend='cpu')
        self.fast_fk2 = jit(self.compute_fk2, backend='cpu')

        print('compiled!')

    def retarget(self, fingertip_1b_se3, fingertip_2b_se3, fingertip_1_se3, fingertip_2_se3, return_cost=False, close_joints_coefficient=1e-2):
        """
        r_h: vector of fingertip targets
        """

        # print(r_h)

        res = minimize(self.fast_c,
                       self.start_angles, method="SLSQP", args=(fingertip_1b_se3, fingertip_2b_se3, fingertip_1_se3, fingertip_2_se3, self.prev_angles,
                                                                close_joints_coefficient),
                       jac=self.grad_c, tol=1e-10, options={'maxiter': 3000}, bounds=self.bounds,
                       )
        self.prev_angles = res.x
        self.start_angles = res.x

        #print("arm retargeter")
        #print("optimal value")
        #print(self.start_angles)

        if return_cost:
            return res.x, res.fun  # ,  [fk1[:3,-1], fk2[:3,-1], fk3[:3,-1], fk4[:3,-1]]
        else:
            return res.x


class FR3AgileHandOptIK:
    def __init__(self, chilarity="right", with_hand = False):
        
        self.joint_names = ["panda_joint1",
                            "panda_joint2",
                            "panda_joint3",
                            "panda_joint4",
                            "panda_joint5",
                            "panda_joint6",
                            "panda_joint7",

                            "right_thumb_0", 
                            "right_thumb_1",
                            "right_thumb_2",
                            "right_thumb_3",

                            "right_index_0",
                            "right_index_1",
                            "right_index_2",
                            "right_index_3",

                            "right_middle_0",
                            "right_middle_1",
                            "right_middle_2",
                            "right_middle_3",

                            "right_ring_0",
                            "right_ring_1",
                            "right_ring_2",
                            "right_ring_3",

                            "right_little_0",
                            "right_little_1",
                            "right_little_2",
                            "right_little_3",

                            ]
        # urdf_path = panda_description.URDF_PATH
        # copy this urdf to ./assets/panda.urdf
        # import shutil 
        # shutil.copy(urdf_path, os.path.join(CUR_PATH, '../assets', 'panda.urdf'))
        urdf_path = os.path.join(CUR_PATH, '../assets', 'fr3_agilehand', 'fr3_agilehand.urdf')
        kinDyn = KinDynComputations(urdf_path,
                                    self.joint_names, 'panda_link0')

        self.fk_wrist = kinDyn.forward_kinematics_fun('panda_link8')
        self.fk_thumb_axis = kinDyn.forward_kinematics_fun('right_thumb_phadist_axis')
        self.fk_index_axis = kinDyn.forward_kinematics_fun('right_index_phadist_axis')
        self.fk_middle_axis = kinDyn.forward_kinematics_fun('right_middle_phadist_axis')
        self.fk_ring_axis = kinDyn.forward_kinematics_fun('right_ring_phadist_axis')
        self.fk_little_axis = kinDyn.forward_kinematics_fun('right_little_phadist_axis')

        lb = [-2.74, -1.78, -2.9, -3.040, -2.81, +0.544, -3.02] \
             + [-0.262, 0, 0, 0] * 5
        ub = [+2.74, +1.78, +2.9, -0.152, +2.81, +4.520, +3.02] \
             + [+0.262, 1.48, 1.13, 1.13] * 5 

        self.bounds = Bounds(lb=lb,
                            ub=ub)

        self.num_joints = len(self.joint_names)

        self.init_start_angles = np.array([0, 0, 0, -1.5707899999999999, 0, 1.5707899999999999, -0.7853] + [0.0] * 20)  

        self.start_angles = np.array([0, 0, 0, -1.5707899999999999, 0, 1.5707899999999999, -0.7853] + [0.0] * 20)  
        self.prev_angles = self.start_angles

        self.home_angles = np.array([0, 0, 0, -1.5707899999999999, 0, 1.5707899999999999, -0.7853] + [0.0] * 20)  

        self.compile()


    def cost(self, angles, wrist, thumb, index, middle, ring, little, prev_angles, close_joints_coefficient):
        H_b = jnp.eye(4)

        tf_w_wrist = self.fk_wrist(H_b, angles)
        tf_w_thumb = self.fk_thumb_axis(H_b, angles)
        tf_w_index = self.fk_index_axis(H_b, angles)
        tf_w_middle = self.fk_middle_axis(H_b, angles)
        tf_w_ring = self.fk_ring_axis(H_b, angles)
        tf_w_little = self.fk_little_axis(H_b, angles)

        cost = 0

        # tf_w_rot1 = tf_w1[:3, :3]
        # tf_w_rot2 = tf_w2[:3, :3]
        # r_h_rot1 = fingertip_1[:3, :3]
        # r_h_rot2 = fingertip_2[:3, :3]

        # rot_error1 = tf_w_rot1.flatten() - r_h_rot1.flatten()
        # rot_error2 = tf_w_rot2.flatten() - r_h_rot2.flatten()
        # cost += jnp.dot(rot_error1, rot_error1) * 1e-5
        # cost += jnp.dot(rot_error2, rot_error2) * 1e-5

        tf_w1 = tf_w_wrist[:3, 3]
        tf_w2 = tf_w_thumb[:3, 3]
        tf_w3 = tf_w_index[:3, 3]
        tf_w4 = tf_w_middle[:3, 3]
        tf_w5 = tf_w_ring[:3, 3]
        tf_w6 = tf_w_little[:3, 3]

        rh_p1 = wrist[:3, 3]
        rh_p2 = thumb[:3, 3]
        rh_p3 = index[:3, 3]
        rh_p4 = middle[:3, 3]
        rh_p5 = ring[:3, 3]
        rh_p6 = little[:3, 3]

        tf_rot = tf_w_wrist[:3, :3]
        rh_rot = wrist[:3, :3]

        rot_error = tf_rot.flatten() - rh_rot.flatten()
        cost += jnp.dot(rot_error, rot_error) * 0.0


        pose_error1 = tf_w1.flatten() - rh_p1.flatten()
        pose_error2 = tf_w2.flatten() - rh_p2.flatten()
        pose_error3 = tf_w3.flatten() - rh_p3.flatten()
        pose_error4 = tf_w4.flatten() - rh_p4.flatten()
        pose_error5 = tf_w5.flatten() - rh_p5.flatten()
        pose_error6 = tf_w6.flatten() - rh_p6.flatten()

        cost += jnp.dot(pose_error1, pose_error1) * 1.0
        # cost += jnp.dot(pose_error2, pose_error2) * 1.0
        # cost += jnp.dot(pose_error3, pose_error3) * 1.0  #e-2
        # cost += jnp.dot(pose_error4, pose_error4) * 1.0 #e-2
        # cost += jnp.dot(pose_error5, pose_error5) * 1.0
        # cost += jnp.dot(pose_error6, pose_error6) * 1.0

        angle_error = angles - prev_angles
        cost += close_joints_coefficient * jnp.dot(angle_error, angle_error)

        # dist_to_home_angle = angles - self.home_angles
        # cost += 1e-5 * jnp.dot(dist_to_home_angle, dist_to_home_angle)

        return cost

    def compile(self):
        self.fast_c = jit(self.cost, backend='cpu')
        self.grad_c = jit(grad(self.fast_c), backend='cpu')

        print('compiled!')

    def retarget(self, wrist, thumb, index, middle, ring, little, return_cost=False, close_joints_coefficient=1e-2):
        """
        r_h: vector of fingertip targets
        """

        # print(r_h)

        res = minimize(self.fast_c,
                       self.start_angles, method="SLSQP", args=(wrist, thumb, index, middle, ring, little, self.prev_angles,
                                                                close_joints_coefficient),
                       jac=self.grad_c, tol=1e-10, options={'maxiter': 6000}, bounds=self.bounds,
                       )
        self.prev_angles = res.x
        self.start_angles = res.x

        #print("arm retargeter")
        #print("optimal value")
        #print(self.start_angles)

        if return_cost:
            return res.x, res.fun  # ,  [fk1[:3,-1], fk2[:3,-1], fk3[:3,-1], fk4[:3,-1]]
        else:
            return res.x



class G1OptIK:
    def __init__(self):
        
        self.joint_names = ["torso_joint",
                            
                            "left_shoulder_pitch_joint",
                            "left_shoulder_roll_joint",
                            "left_shoulder_yaw_joint",
                            "left_elbow_pitch_joint",
                            "left_elbow_roll_joint",

                            "right_shoulder_pitch_joint",
                            "right_shoulder_roll_joint",
                            "right_shoulder_yaw_joint",
                            "right_elbow_pitch_joint",
                            "right_elbow_roll_joint",

                            ]
        # urdf_path = panda_description.URDF_PATH
        # copy this urdf to ./assets/panda.urdf
        # import shutil 
        # shutil.copy(urdf_path, os.path.join(CUR_PATH, '../assets', 'panda.urdf'))
        urdf_path = os.path.join(CUR_PATH, '../assets', 'unitree_g1', 'g1.urdf')
        kinDyn = KinDynComputations(urdf_path,
                                    self.joint_names, 'pelvis')

        self.right_wrist_fk = kinDyn.forward_kinematics_fun('right_wrist_link')
        self.left_wrist_fk = kinDyn.forward_kinematics_fun('left_wrist_link')

        lb = [-np.pi] * 11 #[-2.62] + [-2.97, -1.59, -2.62, -0.227, -2.09, -0.524] # [-0.524, -1, 0, -1.84, -1.84, -1.84, -1.84]
        ub = [+np.pi] * 11 #[+2.74, +1.78, +2.9, -0.152, +2.81, +4.520, +3.02] \
             #+ [+0.262, 1.48, 1.13, 1.13] * 5 

        self.bounds = Bounds(lb=lb,
                            ub=ub)

        self.num_joints = len(self.joint_names)

        self.init_start_angles = np.array([0.0] * 11 )  

        self.start_angles = np.array([0.0] * 11 ) 
        self.prev_angles = self.start_angles

        self.home_angles = np.array([0.0] * 11 ) 

        self.compile()


    def cost(self, angles, right_wrist, left_wrist, prev_angles, close_joints_coefficient):
        H_b = jnp.eye(4)

        tf_w_right_wrist = self.right_wrist_fk(H_b, angles)
        tf_w_left_wrist = self.left_wrist_fk(H_b, angles)

        cost = 0

        # tf_w_rot1 = tf_w1[:3, :3]
        # tf_w_rot2 = tf_w2[:3, :3]
        # r_h_rot1 = fingertip_1[:3, :3]
        # r_h_rot2 = fingertip_2[:3, :3]

        # rot_error1 = tf_w_rot1.flatten() - r_h_rot1.flatten()
        # rot_error2 = tf_w_rot2.flatten() - r_h_rot2.flatten()
        # cost += jnp.dot(rot_error1, rot_error1) * 1e-5
        # cost += jnp.dot(rot_error2, rot_error2) * 1e-5

        right_tf_w1 = tf_w_right_wrist[:3, 3]
        right_rh_p1 = right_wrist[:3, 3]

        right_tf_rot = tf_w_right_wrist[:3, :3]
        right_rh_rot = right_wrist[:3, :3]

        right_rot_error = right_tf_rot.flatten() - right_rh_rot.flatten()
        right_pose_error = right_tf_w1.flatten() - right_rh_p1.flatten()

        cost += jnp.dot(right_rot_error, right_rot_error) * 0.0
        cost += jnp.dot(right_pose_error, right_pose_error) * 1.0

        left_tf_w1 = tf_w_left_wrist[:3, 3]
        left_rh_p1 = left_wrist[:3, 3]

        left_tf_rot = tf_w_left_wrist[:3, :3]
        left_rh_rot = left_wrist[:3, :3]

        left_rot_error = left_tf_rot.flatten() - left_rh_rot.flatten()
        left_pose_error = left_tf_w1.flatten() - left_rh_p1.flatten()

        cost += jnp.dot(left_rot_error, left_rot_error) * 0.0
        cost += jnp.dot(left_pose_error, left_pose_error) * 1.0

        angle_error = angles - prev_angles
        cost += close_joints_coefficient * jnp.dot(angle_error, angle_error)
        
        dist_to_home_angle = angles - self.home_angles
        cost += 1e-3 * jnp.dot(dist_to_home_angle, dist_to_home_angle)

        return cost

    def compile(self):
        self.fast_c = jit(self.cost, backend='cpu')
        self.grad_c = jit(grad(self.fast_c), backend='cpu')

        print('compiled!')

    def retarget(self, right_wrist, left_wrist, return_cost=False, close_joints_coefficient=1e-2):
        """
        r_h: vector of fingertip targets
        """

        # print(r_h)

        res = minimize(self.fast_c,
                       self.start_angles, method="SLSQP", args=(right_wrist, left_wrist, self.prev_angles,
                                                                close_joints_coefficient),
                       jac=self.grad_c, tol=1e-10, options={'maxiter': 6000}, bounds=self.bounds,
                       )
        self.prev_angles = res.x
        self.start_angles = res.x

        #print("arm retargeter")
        #print("optimal value")
        #print(self.start_angles)

        if return_cost:
            return res.x, res.fun  # ,  [fk1[:3,-1], fk2[:3,-1], fk3[:3,-1], fk4[:3,-1]]
        else:
            return res.x




# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

import dart_physics.avp_stream3.grpc_msg.handtracking_pb2 as handtracking__pb2


class HandTrackingServiceStub(object):
    """The hand tracking service definition.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.StreamHandUpdates = channel.unary_stream(
                '/handtracking.HandTrackingService/StreamHandUpdates',
                request_serializer=handtracking__pb2.HandUpdate.SerializeToString,
                response_deserializer=handtracking__pb2.HandUpdate.FromString,
                )


class HandTrackingServiceServicer(object):
    """The hand tracking service definition.
    """

    def StreamHandUpdates(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_HandTrackingServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'StreamHandUpdates': grpc.unary_stream_rpc_method_handler(
                    servicer.StreamHandUpdates,
                    request_deserializer=handtracking__pb2.HandUpdate.FromString,
                    response_serializer=handtracking__pb2.HandUpdate.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'handtracking.HandTrackingService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class HandTrackingService(object):
    """The hand tracking service definition.
    """

    @staticmethod
    def StreamHandUpdates(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/handtracking.HandTrackingService/StreamHandUpdates',
            handtracking__pb2.HandUpdate.SerializeToString,
            handtracking__pb2.HandUpdate.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)


class VideoStreamStub(object):
    """The video stream service definition
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.SendMatrices = channel.unary_unary(
                '/handtracking.VideoStream/SendMatrices',
                request_serializer=handtracking__pb2.Matrices.SerializeToString,
                response_deserializer=handtracking__pb2.StreamAck.FromString,
                )


class VideoStreamServicer(object):
    """The video stream service definition
    """

    def SendMatrices(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_VideoStreamServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'SendMatrices': grpc.unary_unary_rpc_method_handler(
                    servicer.SendMatrices,
                    request_deserializer=handtracking__pb2.Matrices.FromString,
                    response_serializer=handtracking__pb2.StreamAck.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'handtracking.VideoStream', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class VideoStream(object):
    """The video stream service definition
    """

    @staticmethod
    def SendMatrices(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/handtracking.VideoStream/SendMatrices',
            handtracking__pb2.Matrices.SerializeToString,
            handtracking__pb2.StreamAck.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)


class ControlServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.SendControlCommand = channel.unary_unary(
                '/handtracking.ControlService/SendControlCommand',
                request_serializer=handtracking__pb2.ControlCommand.SerializeToString,
                response_deserializer=handtracking__pb2.ControlResponse.FromString,
                )


class ControlServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def SendControlCommand(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_ControlServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'SendControlCommand': grpc.unary_unary_rpc_method_handler(
                    servicer.SendControlCommand,
                    request_deserializer=handtracking__pb2.ControlCommand.FromString,
                    response_serializer=handtracking__pb2.ControlResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'handtracking.ControlService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class ControlService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def SendControlCommand(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/handtracking.ControlService/SendControlCommand',
            handtracking__pb2.ControlCommand.SerializeToString,
            handtracking__pb2.ControlResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



import meshcat
import meshcat.geometry as g
import meshcat.transformations as tf
import os
import mujoco 
from scipy.spatial.transform import Rotation as R
import numpy as np 
from meshcat.animation import Animation
from dm_control import mjcf
from io import BytesIO

class MJCMeshcat: 

    def __init__(self): 

        # Create a new visualizer
        self.vis = meshcat.Visualizer()
        self.base_t = dict() 
        self.materials = dict() 
        self.meshes = dict()

        self.anim = Animation()
        self.cnt = 0 

    def parse_mjcf(self, mjcf: mjcf.Element): 

        meshcat_dict = {} 

        meshcat_dict["material"] = {}
        for material in mjcf.asset.find_all("material"):
            name = material.full_identifier 
            rgba = material.rgba
            texture = material.texture

            if texture is not None and texture != "grid": 
                texture_contents = BytesIO(texture.file.contents)
            else: 
                texture_contents = None
            
            self.register_material(name, texture_contents, rgba)

        for mesh in mjcf.asset.find_all("mesh"):

            input_io = BytesIO(mesh.file.contents)

            try: 
                scale = mesh.scale[0]
            except:
                scale = 1.0

            # print(mesh.name, mesh.file.extension, scale)
            self.register_mesh(mesh.name, mesh.file.extension, input_io, scale)


        for body in mjcf.worldbody.find_all("body"):
            if body.geom is None: 
                continue
            for geom in body.geom:
                # print(dir(geom))
                # print(geom.__class__)
                # for method in dir(geom):
                #     print(method, ':', str(getattr(geom, method)))
                if "collision" in str(geom.dclass): 
                    pass
                else:
                    try: 
                        if geom.material is None: 
                            # print(str(geom.dclass))
                            material = geom.dclass.geom.material 
                        else:
                            material = geom.material
                        self.register_geom(body.full_identifier, geom.mesh.name, material.full_identifier, geom.pos, geom.quat)
                    except Exception as e:
                        # print(body.full_identifier, geom.name, e)
                        pass 


    def register_material(self, material_name, texture_content, rgba):


        if texture_content is None and rgba is not None: 
            # convert rgba to hex like 0xffffff
            import matplotlib.colors as mcolors
            hex = mcolors.to_hex(rgba)
            hex = hex.replace("#", "0x")

            self.materials[material_name] = g.MeshLambertMaterial(color=hex)

        elif texture_content is not None and rgba is None:
            self.materials[material_name] = g.MeshLambertMaterial(
                map=g.ImageTexture(g.PngImage.from_stream(texture_content))
            )


    def register_mesh(self, mesh_name, mesh_type, mesh_content, mesh_scale = 1.0):

        # Load the mesh
        if "stl" in mesh_type: 
            mesh = g.StlMeshGeometry.from_stream(mesh_content)
        elif "obj" in mesh_type:
            mesh = g.ObjMeshGeometry.from_stream(mesh_content)


        self.meshes[mesh_name] = {
            "mesh": mesh,
            "scale": mesh_scale, 
        }


        
    def register_geom(self, body_name, mesh_name, material_name, pos, quat):
        

        # Get the mesh
        mesh = self.meshes[mesh_name]["mesh"]
        scale = self.meshes[mesh_name]["scale"]

        # Get the material
        if material_name not in self.materials: 
            material = None 
            
        else:
            material = self.materials[material_name]

        # Get the position and orientation
        mat = np.eye(4)
        if quat is not None:    
            mat[:3, :3] = R.from_quat(quat, scalar_first=True).as_matrix()
        if pos is not None: 
            mat[:3, -1] = pos
        
        mat[:3, :3] *= scale

        self.base_t[f"{body_name}/{mesh_name}"] = mat


        # Create the geometry

        self.vis[f"{body_name}/{mesh_name}"].set_object(mesh, material)
        # geom.set_transform(tf.quaternion_matrix(quat))
        self.vis[f"{body_name}/{mesh_name}"].set_transform(self.base_t[f"{body_name}/{mesh_name}"])

        # # Add the geometry to the visualizer
        print(f"Added {body_name}/{mesh_name} to the visualizer")


    def set_body_transform(self, body_name, pos, quat, frame): 

        if frame is None: 
            frame = self.vis 

        mat = np.eye(4)
        if quat is not None:    
            mat[:3, :3] = R.from_quat(quat, scalar_first=True).as_matrix()
        if pos is not None: 
            mat[:3, -1] = pos

        frame[body_name].set_transform(mat)


    @property
    def bodies(self): 
        # return the list of bodies
        # you can extract it from self.vis keys which are {body_name}/{geom_name}

        d =  [] 
        for key in self.base_t.keys(): 
            c1, c2, c3 = key.split("/")
            d.append(f"{c1}/{c2}")

        return d
        

    def set_keypoints(self, body_transforms):

        with self.anim.at_frame(self.vis, self.cnt/10.) as frame:
            for transform in body_transforms: 
                body_name, (pos, quat) = transform
                self.set_body_transform(body_name, pos, quat, frame)

        self.cnt += 1

        
    def get_html(self, path): 
        self.vis.set_animation(self.anim, play=True, repetitions = 100) 

        html = self.vis.static_html()
        return html 
        

if __name__ == "__main__":

    mjc = MJCMeshcat()


"""Damping task implementation."""

from __future__ import annotations

import mujoco
import numpy.typing as npt

from .posture_task import PostureTask


class DampingTask(PostureTask):
    """Minimize joint velocities.

    This task is implemented as a special case of the PostureTask where the gain and
    target configuration are set to 0 and qpos0 respectively.
    """

    def __init__(self, model: mujoco.MjModel, cost: npt.ArrayLike):
        super().__init__(model=model, cost=cost, gain=0.0, lm_damping=0.0)
        self.target_q = model.qpos0




from dart_physics.robot_cfgs import * 
from pathlib import Path
import dart_physics.mink as mink 

_HERE = Path(__file__).resolve().parent


def qpos2ctrl(model, data, qpos_target): 

    ctrl = np.zeros(16)
    ctrl[:7] = qpos_target[0:7]
    ctrl[8:15] = qpos_target[9:16]

    # left gripper control
    tip1 = mink.SE3.from_mocap_name(model, data, "left_lf_target").as_matrix()
    tip2 = mink.SE3.from_mocap_name(model, data, "left_rf_target").as_matrix()

    finger_dist = np.linalg.norm(tip1[:3, 3] - tip2[:3, 3])
    max_grip = data.qpos[8] - 0.01
    ctrl[7] = np.clip(finger_dist - 0.01, max_grip, 0.04)

    # right gripper control
    tip1 = mink.SE3.from_mocap_name(model, data, "right_lf_target").as_matrix()
    tip2 = mink.SE3.from_mocap_name(model, data, "right_rf_target").as_matrix()

    finger_dist = np.linalg.norm(tip1[:3, 3] - tip2[:3, 3])
    max_grip = data.qpos[16] - 0.01
    ctrl[15] = np.clip(finger_dist - 0.01, max_grip, 0.04)


    return ctrl 



robot_cfg = {

    "name": "DualPanda",

    "robots": { 
        "l_robot": RobotConfig( 
                    xml_path = (_HERE / "../assets" / "franka_emika_panda" / "new_panda.xml").as_posix(),
                    home_q = [0, 0, 0, -1.5707899999999999, 0, 1.5707899999999999, -0.7853, 0.040000000000000001, 0.040000000000000001], 
                    freejoint = False, 
                    attach_to = [-0.4, -0.15, 0.0, 0.7071068 , 0, 0, 0.7071068 ], 
                    parallel_jaw = True, 
                    add_bodies = [ 
                        NewBodyConfig(attach_to = "right_finger", \
                                      body_name = "right_finger_base", pos = [0.0, 0.02, 0.0]),
                        NewBodyConfig(attach_to = "right_finger", \
                                        body_name = "right_finger_tip", pos = [0, 0, 0.05]), 
                        NewBodyConfig(attach_to = "left_finger", \
                                        body_name = "left_finger_base", pos = [0.0, 0.02, 0.0]),
                        NewBodyConfig(attach_to = "left_finger", \
                                        body_name = "left_finger_tip", pos = [0, 0, 0.05]),
                    ], 
                    
                ), 
        "r_robot": RobotConfig(
                    xml_path = (_HERE / "../assets" / "franka_emika_panda" / "new_panda.xml").as_posix(),
                    home_q = [0, 0, 0, -1.5707899999999999, 0, 1.5707899999999999, -0.7853, 0.040000000000000001, 0.040000000000000001], 
                    freejoint = False,
                    attach_to = [0.4, -0.15, 0.0, 0.7071068 , 0, 0, 0.7071068 ], 
                    parallel_jaw = True,
                    add_bodies = [ 
                        NewBodyConfig(attach_to = "right_finger", \
                                      body_name = "right_finger_base", pos = [0.0, 0.02, 0.0]),
                        NewBodyConfig(attach_to = "right_finger", \
                                        body_name = "right_finger_tip", pos = [0, 0, 0.05]), 
                        NewBodyConfig(attach_to = "left_finger", \
                                        body_name = "left_finger_base", pos = [0.0, 0.02, 0.0]),
                        NewBodyConfig(attach_to = "left_finger", \
                                        body_name = "left_finger_tip", pos = [0, 0, 0.05]),
                    ], 

                ) 
    }, 

    "qpos2ctrl": qpos2ctrl, 
    "obj_startidx": (7+2) * 2, 
        
    "points" : { 

        "left_lb":     FingerKeypoints("left", 3, "left_lb_target", "l_robot/left_finger_base", type = "body"),
        "left_lf":      FingerKeypoints("left", 4, "left_lf_target", "l_robot/left_finger_tip", type = "body"),
        "left_rb":      FingerKeypoints("left", 8, "left_rb_target", "l_robot/right_finger_base", type = "body"),
        "left_rf":      FingerKeypoints("left", 9, "left_rf_target", "l_robot/right_finger_tip", type = "body"),
        
        "right_lb":     FingerKeypoints("right", 3, "right_lb_target", "r_robot/left_finger_base", type = "body"),
        "right_lf":     FingerKeypoints("right", 4, "right_lf_target", "r_robot/left_finger_tip", type = "body"),
        "right_rb":     FingerKeypoints("right", 8, "right_rb_target", "r_robot/right_finger_base", type = "body"),
        "right_rf":     FingerKeypoints("right", 9, "right_rf_target", "r_robot/right_finger_tip", type = "body"),

    },

    "reverse_points": { 

        "left_lb":     FingerKeypoints("right", 3, "left_lb_target", "l_robot/right_finger_base", type = "body"),
        "left_lf":      FingerKeypoints("right", 4, "left_lf_target", "l_robot/right_finger_tip", type = "body"),
        "left_rb":      FingerKeypoints("right", 8, "left_rb_target", "l_robot/left_finger_base", type = "body"),
        "left_rf":      FingerKeypoints("right", 9, "left_rf_target", "l_robot/left_finger_tip", type = "body"),
        
        "right_lb":     FingerKeypoints("left", 3, "right_lb_target", "r_robot/right_finger_base", type = "body"),
        "right_lf":     FingerKeypoints("left", 4, "right_lf_target", "r_robot/right_finger_tip", type = "body"),
        "right_rb":     FingerKeypoints("left", 8, "right_rb_target", "r_robot/left_finger_base", type = "body"),
        "right_rf":     FingerKeypoints("left", 9, "right_rf_target", "r_robot/left_finger_tip", type = "body"),

    }, 

    "avp_calib": { 

        "left_lb": {"scale": 1.0, "offset": np.array([0.0, 0.0, 0.0])},
        "left_lf": {"scale": 1.0, "offset": np.array([0.0, 0.0, 0.0])},
        "left_rb": {"scale": 1.0, "offset": np.array([0.0, 0.0, 0.0])},
        "left_rf": {"scale": 1.0, "offset": np.array([0.0, 0.0, 0.0])},

        "right_lb": {"scale": 1.0, "offset": np.array([0.0, 0.0, 0.0])},
        "right_lf": {"scale": 1.0, "offset": np.array([0.0, 0.0, 0.0])},
        "right_rb": {"scale": 1.0, "offset": np.array([0.0, 0.0, 0.0])},
        "right_rf": {"scale": 1.0, "offset": np.array([0.0, 0.0, 0.0])},

    }, 

    "ik_task" : [
        IKTasks(target = "left_lb"), 
        IKTasks(target = "left_lf"),
        IKTasks(target = "left_rb"),
        IKTasks(target = "left_rf"),

        IKTasks(target = "right_lb"),
        IKTasks(target = "right_lf"),
        IKTasks(target = "right_rb"),
        IKTasks(target = "right_rf"),
    ],

    "joints" : ['l_robot/joint{}'.format(i) for i in range(1, 8)] + ['r_robot/joint{}'.format(i) for i in range(1, 8)], 

    "bodies":  {
        # left robot
        "l_robot/link0":   "panda_link0_left",   "l_robot/link1":   "panda_link1_left", 
        "l_robot/link2":   "panda_link2_left",   "l_robot/link3":   "panda_link3_left",
        "l_robot/link4":   "panda_link4_left",   "l_robot/link5":   "panda_link5_left",
        "l_robot/link6":   "panda_link6_left",   "l_robot/link7":   "panda_link7_left",
        "l_robot/hand":    "panda_hand_left",    "l_robot/left_finger": "panda_left_finger_left", 
        "l_robot/right_finger": "panda_right_finger_left",


        # right robot
        "r_robot/link0":   "panda_link0_right",   "r_robot/link1":   "panda_link1_right",
        "r_robot/link2":   "panda_link2_right",   "r_robot/link3":   "panda_link3_right",
        "r_robot/link4":   "panda_link4_right",   "r_robot/link5":   "panda_link5_right",
        "r_robot/link6":   "panda_link6_right",   "r_robot/link7":   "panda_link7_right",
        "r_robot/hand":    "panda_hand_right",    "r_robot/left_finger": "panda_left_finger_right",
        "r_robot/right_finger": "panda_right_finger_right",

    } 

} 


if __name__ == "__main__": 

    import trimesh 
    from scipy.spatial.transform import Rotation as R
    import os 

    asset_root = os.path.join(_HERE, "../assets", "aloha", "assets")
    usdz_root = os.path.join(_HERE, "../assets", "aloha", "transformed_stl")\

    os.makedirs(usdz_root, exist_ok = True)

    print([f"{k}" for k in robot_cfg["bodies"].values()])

    for k, v in robot_cfg["usdz_geoms"].items():

        meshes = [] 
        for i, geom in enumerate(v): 
            geom: GeomInfo
            mesh = trimesh.load_mesh(f"{asset_root}/{geom.mesh}.stl")
            pos = np.array(geom.pos)
            quat = np.array(geom.quat)

            mesh.visual.vertex_colors = np.array([0, 0, 0, 255], dtype = np.uint8)

            mat = np.eye(4)
            mat[:3, :3] = R.from_quat(quat, scalar_first = True).as_matrix()
            mat[:3, 3] = pos

            mesh.apply_transform(mat)

            meshes.append(mesh)

        scale = [mesh.scale for mesh in meshes]
        avg_scale = np.mean(scale, axis = 0)
        print(k, avg_scale)
        
        meshes = trimesh.util.concatenate(meshes)

        if avg_scale > 100: 

            meshes.apply_scale(0.001)

        if avg_scale < 1: 

            meshes.apply_scale(1)
        
        meshes.export(f"{usdz_root}/{k}.glb")





import sys
from dart_physics.robot_cfgs import GeomInfo
from scipy.spatial.transform import Rotation as R
import random
from pathlib import Path
from dart_physics.utils.run_functions import * 

HERE = Path(__file__).parent.absolute()


def random_xyz(i, xmin = -0.2, xmax= 0.2, ymin = 0.2, ymax = 0.4, zmin = 0.1):
    return [random.uniform(xmin, xmax), random.uniform(ymin, ymax), zmin + 0.03 * i ]

def duplicate_objects(name, num_max): 

    objs =  [name + "_{}".format(i) if i>0 else name for i in range(num_max)]

    # replicate the object 
    import shutil 
    obj_path = f"{HERE}/../assets/custom/{name}/{name}"
    
    for i in range(len(objs)):
        copy_path = f"{HERE}/../assets/custom/{name}_{i}"
        try: 
            shutil.copytree(obj_path, copy_path)
            shutil.move(f"{copy_path}/{name}.xml", f"{copy_path}/{name}_{i}.xml")
            # change the model name in the xml file
        except: 
            pass 
        # change the xml file (name.xml) to (name{i}.xml)

    return objs

def reset_function(model, data, robot_cfg, task_cfg): 

    id = robot_cfg["obj_startidx"] 
    vid = robot_cfg["obj_startidx"] 

    for ii, obj in enumerate(task_cfg['objects']): 

        if obj == "toy_five_insertion_base": 

            new_pos = task_cfg["default_poses"][obj][:3]  
            new_quat = R.from_euler('xyz', [0, 0, -30 + np.random.randint(60)], degrees = True).as_quat(scalar_first = True).tolist()

        else:
            new_pos = task_cfg["default_poses"][obj][:3]
            new_quat = task_cfg["default_poses"][obj][3:7]

        print(obj, len(new_pos))
        data.qpos[id:id+3] = new_pos
        data.qpos[id+3:id+7] = new_quat
        data.qvel[vid:vid+6] = 0.0

        id += 7 
        vid += 6


jengas = duplicate_objects("jenga", 24)

task_cfg = { 
    "name": "toy_five_insertions", 

    "objects": jengas, 

    "cone": "elliptic", 
    "o_solref": ".000001 1",
    "o_solimp": ".001 1",
    "o_friction": "0.1 0.1 1", 
    "impratio": "100.0",
    "multiccd": "disable" ,
    "override": "enable",

    "default_poses": {
        
        "jenga": [-0.026, 0.2, 0.03] + R.from_euler('xyz', [0, 0, 90], degrees = True).as_quat(scalar_first = True).tolist(), 
        "jenga_1": [0.026, 0.2, 0.03] + R.from_euler('xyz', [0, 0, 90], degrees = True).as_quat(scalar_first = True).tolist(), 
        "jenga_2": [0.0, 0.2, 0.03] + R.from_euler('xyz', [0, 0, 90], degrees = True).as_quat(scalar_first = True).tolist(), 
        "jenga_3": [0.0, 0.2 - 0.026, 0.046] + R.from_euler('xyz', [0, 0, 0], degrees = True).as_quat(scalar_first = True).tolist(), 
        "jenga_4": [0.0, 0.2 + 0.026, 0.046] + R.from_euler('xyz', [0, 0, 0], degrees = True).as_quat(scalar_first = True).tolist(), 
        "jenga_5": [0.0, 0.2, 0.046] + R.from_euler('xyz', [0, 0, 0], degrees = True).as_quat(scalar_first = True).tolist(), 
        "jenga_6": [-0.026, 0.2, 0.062] + R.from_euler('xyz', [0, 0, 90], degrees = True).as_quat(scalar_first = True).tolist(), 
        "jenga_7": [0.026, 0.2, 0.062] + R.from_euler('xyz', [0, 0, 90], degrees = True).as_quat(scalar_first = True).tolist(), 
        "jenga_8": [0.0, 0.2, 0.062] + R.from_euler('xyz', [0, 0, 90], degrees = True).as_quat(scalar_first = True).tolist(), 
        
        "jenga_9": [0.0, 0.2-0.026, 0.078] + R.from_euler('xyz', [0, 0, 0], degrees = True).as_quat(scalar_first = True).tolist(), 
        "jenga_10": [0.0, 0.2 + 0.026, 0.078] + R.from_euler('xyz', [0, 0, 0], degrees = True).as_quat(scalar_first = True).tolist(), 
        "jenga_11": [0.0, 0.2, 0.078] + R.from_euler('xyz', [0, 0, 0], degrees = True).as_quat(scalar_first = True).tolist(), 

        "jenga_12": [0.026, 0.2, 0.094] + R.from_euler('xyz', [0, 0, 90], degrees = True).as_quat(scalar_first = True).tolist(), 
        "jenga_13": [-0.026, 0.2, 0.094] + R.from_euler('xyz', [0, 0, 90], degrees = True).as_quat(scalar_first = True).tolist(), 
        "jenga_14": [0.0, 0.2, 0.094] + R.from_euler('xyz', [0, 0, 90], degrees = True).as_quat(scalar_first = True).tolist(), 

        "jenga_15": [0.0, 0.2-0.026, 0.11] + R.from_euler('xyz', [0, 0, 0], degrees = True).as_quat(scalar_first = True).tolist(), 
        "jenga_16": [0.0, 0.2 + 0.026, 0.11] + R.from_euler('xyz', [0, 0, 0], degrees = True).as_quat(scalar_first = True).tolist(), 
        "jenga_17": [0.0, 0.2, 0.11] + R.from_euler('xyz', [0, 0, 0], degrees = True).as_quat(scalar_first = True).tolist(), 

        "jenga_18": [0.026, 0.2, 0.126] + R.from_euler('xyz', [0, 0, 90], degrees = True).as_quat(scalar_first = True).tolist(), 
        "jenga_19": [-0.026, 0.2, 0.126] + R.from_euler('xyz', [0, 0, 90], degrees = True).as_quat(scalar_first = True).tolist(), 
        "jenga_20": [0.0, 0.2, 0.126] + R.from_euler('xyz', [0, 0, 90], degrees = True).as_quat(scalar_first = True).tolist(), 

        "jenga_21": [0.0, 0.2-0.026, 0.142] + R.from_euler('xyz', [0, 0, 0], degrees = True).as_quat(scalar_first = True).tolist(), 
        "jenga_22": [0.0, 0.2 + 0.026, 0.142] + R.from_euler('xyz', [0, 0, 0], degrees = True).as_quat(scalar_first = True).tolist(), 
        "jenga_23": [0.0, 0.2, 0.142] + R.from_euler('xyz', [0, 0, 0], degrees = True).as_quat(scalar_first = True).tolist(), 


    #     "toy_five_insertion_base":  [-0.1, 0.35, 0.1 ] + R.from_euler('xyz', [0, 0, 0], degrees = True).as_quat(scalar_first = True).tolist(), 
    #     "toy_five_insertion_one":   [+0.1, 0.35, 0.1 ] + R.from_euler('xyz', [0, 0, 0], degrees = True).as_quat(scalar_first = True).tolist(),
    #     "toy_five_insertion_two":   [+0.07, 0.3, 0.12] + R.from_euler('xyz', [0, 0, 0], degrees = True).as_quat(scalar_first = True).tolist(),
    #     "toy_five_insertion_three": [+0.13, 0.4, 0.14] + R.from_euler('xyz', [0, 0, 0], degrees = True).as_quat(scalar_first = True).tolist(),
    #     "toy_five_insertion_four":  [+0.07, 0.4, 0.16] + R.from_euler('xyz', [0, 0, 0], degrees = True).as_quat(scalar_first = True).tolist(),
    #     "toy_five_insertion_five":  [+0.13, 0.3, 0.18] + R.from_euler('xyz', [0, 0, 0], degrees = True).as_quat(scalar_first = True).tolist(),
    }, 

    "reset_function": reset_function, 
    
}

# task_cfg["default_poses"].update({obj: random_xyz(i) + R.from_euler('xyz', [0, 0, np.random.randint(180)], degrees = True).as_quat(scalar_first = True).tolist() for i, obj in enumerate(jengas)})



if __name__ == "__main__":


    from runs import * 
    from utils.scene_gen import construct_scene
    import mujoco
    from loop_rate_limiters import RateLimiter


    robot = "dual_panda"
    robot_cfg = load_robot_cfg(robot)

    model = construct_scene(task_cfg, robot_cfg)
    data = mujoco.MjData(model)

    viewer = mujoco.viewer.launch_passive(
        model=model, data=data, show_left_ui=True, show_right_ui=True)
    
    mujoco.mj_resetDataKeyframe(model, data, 0)
    mujoco.mj_forward(model, data)
    mujoco.mj_camlight(model, data)

    rate = RateLimiter(500)
    while True:
        mujoco.mj_step(model, data)
        viewer.sync()
        rate.sleep()
    



import sys
from dart_physics.robot_cfgs import GeomInfo
from scipy.spatial.transform import Rotation as R
import random
from pathlib import Path
from dart_physics.utils.run_functions import * 

HERE = Path(__file__).parent.absolute()

def random_xyz(i):
    return [random.uniform(-0.1, 0.1), random.uniform(0.15, 0.2), 0.01 + 0.01 * i ]

def duplicate_objects(name, num_max): 

    objs =  [name + "_{}".format(i) if i>0 else name for i in range(num_max)]

    # replicate the object 
    import shutil 
    obj_path = f"{HERE}/../assets/custom/{name}/{name}"
    
    for i in range(len(objs)):
        copy_path = f"{HERE}/../assets/custom/{name}_{i}"
        try: 
            shutil.copytree(obj_path, copy_path)
            shutil.move(f"{copy_path}/{name}.xml", f"{copy_path}/{name}_{i}.xml")
            # change the model name in the xml file
        except: 
            pass 
        # change the xml file (name.xml) to (name{i}.xml)

    return objs

def reset_function(model, data, robot_cfg, task_cfg): 

    id = robot_cfg["obj_startidx"] 
    vid = robot_cfg["obj_startidx"] 

    table_height = task_cfg.get("table_height", 0.0)

    for ii, obj in enumerate(task_cfg['objects']): 

        if "basket" in obj: 
            pass 

        else:
            new_pos = random_xyz(ii)
            new_quat = R.from_euler('xyz', [0, 0, np.random.randint(180)], degrees = True).as_quat(scalar_first = True).tolist()

            print(obj, len(new_pos))
            data.qpos[id:id+3] = new_pos
            data.qpos[id+3:id+7] = new_quat
            data.qvel[vid:vid+6] = 0.0

            data.qpos[id+2] += table_height

            id += 7 
            vid += 6


utensils = duplicate_objects("sorting_bolt2", 5) + duplicate_objects("sorting_nut1", 5)

task_cfg = { 
    "name": "stack_dishes", 

    "objects": ["wood_basket", "wood_basket_2"] + utensils, 

    "fixed_objects": ["wood_basket", "wood_basket_2"],

    "cone": "elliptic", 
    "o_solref": ".000000001 1",
    "o_solimp": ".001 1",
    "impratio": "100.0",
    "multiccd": "disable" ,
    "override": "enable",

    "default_poses": {
        "wood_basket": [-0.1, 0.35, 0.0] + R.from_euler('xyz', [90, 0, 0], degrees = True).as_quat(scalar_first = True).tolist(), 
        "wood_basket_2": [+0.1, 0.35, 0.0] + R.from_euler('xyz', [90, 0, 0], degrees = True).as_quat(scalar_first = True).tolist(),
    }, 

    "reset_function": reset_function, 
    
}

task_cfg["default_poses"].update({obj: random_xyz(i) + R.from_euler('xyz', [0, 0, np.random.randint(180)], degrees = True).as_quat(scalar_first = True).tolist() for i, obj in enumerate(utensils)})



if __name__ == "__main__":


    from runs import * 
    from utils.scene_gen import construct_scene
    import mujoco
    from loop_rate_limiters import RateLimiter


    robot = "dual_panda"
    robot_cfg = load_robot_cfg(robot)

    model = construct_scene(task_cfg, robot_cfg)
    data = mujoco.MjData(model)

    viewer = mujoco.viewer.launch_passive(
        model=model, data=data, show_left_ui=True, show_right_ui=True)
    
    mujoco.mj_resetDataKeyframe(model, data, 0)
    mujoco.mj_forward(model, data)
    mujoco.mj_camlight(model, data)

    rate = RateLimiter(500)
    while True:
        mujoco.mj_step(model, data)
        viewer.sync()
        rate.sleep()
    


"""Collision avoidance limit."""

import itertools
from dataclasses import dataclass
from typing import List, Sequence, Union

import mujoco
import numpy as np

from ..configuration import Configuration
from .limit import Constraint, Limit

# Type aliases.
Geom = Union[int, str]
GeomSequence = Sequence[Geom]
CollisionPair = tuple[GeomSequence, GeomSequence]
CollisionPairs = Sequence[CollisionPair]


@dataclass(frozen=True)
class Contact:
    dist: float
    fromto: np.ndarray
    geom1: int
    geom2: int
    distmax: float

    @property
    def normal(self) -> np.ndarray:
        normal = self.fromto[3:] - self.fromto[:3]
        return normal / (np.linalg.norm(normal) + 1e-9)

    @property
    def inactive(self) -> bool:
        return self.dist == self.distmax and not self.fromto.any()


def compute_contact_normal_jacobian(
    model: mujoco.MjModel,
    data: mujoco.MjData,
    contact: Contact,
) -> np.ndarray:
    geom1_body = model.geom_bodyid[contact.geom1]
    geom2_body = model.geom_bodyid[contact.geom2]
    geom1_contact_pos = contact.fromto[:3]
    geom2_contact_pos = contact.fromto[3:]
    jac2 = np.empty((3, model.nv))
    mujoco.mj_jac(model, data, jac2, None, geom2_contact_pos, geom2_body)
    jac1 = np.empty((3, model.nv))
    mujoco.mj_jac(model, data, jac1, None, geom1_contact_pos, geom1_body)
    return contact.normal @ (jac2 - jac1)


def _is_welded_together(model: mujoco.MjModel, geom_id1: int, geom_id2: int) -> bool:
    """Returns true if the geoms are part of the same body, or if their bodies are
    welded together."""
    body1 = model.geom_bodyid[geom_id1]
    body2 = model.geom_bodyid[geom_id2]
    weld1 = model.body_weldid[body1]
    weld2 = model.body_weldid[body2]
    return weld1 == weld2


def _are_geom_bodies_parent_child(
    model: mujoco.MjModel, geom_id1: int, geom_id2: int
) -> bool:
    """Returns true if the geom bodies have a parent-child relationship."""
    body_id1 = model.geom_bodyid[geom_id1]
    body_id2 = model.geom_bodyid[geom_id2]

    # body_weldid is the ID of the body's weld.
    body_weldid1 = model.body_weldid[body_id1]
    body_weldid2 = model.body_weldid[body_id2]

    # weld_parent_id is the ID of the parent of the body's weld.
    weld_parent_id1 = model.body_parentid[body_weldid1]
    weld_parent_id2 = model.body_parentid[body_weldid2]

    # weld_parent_weldid is the weld ID of the parent of the body's weld.
    weld_parent_weldid1 = model.body_weldid[weld_parent_id1]
    weld_parent_weldid2 = model.body_weldid[weld_parent_id2]

    cond1 = body_weldid1 == weld_parent_weldid2
    cond2 = body_weldid2 == weld_parent_weldid1
    return cond1 or cond2


def _is_pass_contype_conaffinity_check(
    model: mujoco.MjModel, geom_id1: int, geom_id2: int
) -> bool:
    """Returns true if the geoms pass the contype/conaffinity check."""
    cond1 = bool(model.geom_contype[geom_id1] & model.geom_conaffinity[geom_id2])
    cond2 = bool(model.geom_contype[geom_id2] & model.geom_conaffinity[geom_id1])
    return cond1 or cond2


class CollisionAvoidanceLimit(Limit):
    """Normal velocity limit between geom pairs.

    Attributes:
        model: MuJoCo model.
        geom_pairs: Set of collision pairs in which to perform active collision
            avoidance. A collision pair is defined as a pair of geom groups. A geom
            group is a set of geom names. For each geom pair, the solver will
            attempt to compute joint velocities that avoid collisions between every
            geom in the first geom group with every geom in the second geom group.
            Self collision is achieved by adding a collision pair with the same
            geom group in both pair fields.
        gain: Gain factor in (0, 1] that determines how fast the geoms are
            allowed to move towards each other at each iteration. Smaller values
            are safer but may make the geoms move slower towards each other.
        minimum_distance_from_collisions: The minimum distance to leave between
            any two geoms. A negative distance allows the geoms to penetrate by
            the specified amount.
        collision_detection_distance: The distance between two geoms at which the
            active collision avoidance limit will be active. A large value will
            cause collisions to be detected early, but may incur high computational
            cost. A negative value will cause the geoms to be detected only after
            they penetrate by the specified amount.
        bound_relaxation: An offset on the upper bound of each collision avoidance
            constraint.
    """

    def __init__(
        self,
        model: mujoco.MjModel,
        geom_pairs: CollisionPairs,
        gain: float = 0.85,
        minimum_distance_from_collisions: float = 0.005,
        collision_detection_distance: float = 0.01,
        bound_relaxation: float = 0.0,
    ):
        """Initialize collision avoidance limit.

        Args:
            model: MuJoCo model.
            geom_pairs: Set of collision pairs in which to perform active collision
                avoidance. A collision pair is defined as a pair of geom groups. A geom
                group is a set of geom names. For each collision pair, the mapper will
                attempt to compute joint velocities that avoid collisions between every
                geom in the first geom group with every geom in the second geom group.
                Self collision is achieved by adding a collision pair with the same
                geom group in both pair fields.
            gain: Gain factor in (0, 1] that determines how fast the geoms are
                allowed to move towards each other at each iteration. Smaller values
                are safer but may make the geoms move slower towards each other.
            minimum_distance_from_collisions: The minimum distance to leave between
                any two geoms. A negative distance allows the geoms to penetrate by
                the specified amount.
            collision_detection_distance: The distance between two geoms at which the
                active collision avoidance limit will be active. A large value will
                cause collisions to be detected early, but may incur high computational
                cost. A negative value will cause the geoms to be detected only after
                they penetrate by the specified amount.
            bound_relaxation: An offset on the upper bound of each collision avoidance
                constraint.
        """
        self.model = model
        self.gain = gain
        self.minimum_distance_from_collisions = minimum_distance_from_collisions
        self.collision_detection_distance = collision_detection_distance
        self.bound_relaxation = bound_relaxation
        self.geom_id_pairs = self._construct_geom_id_pairs(geom_pairs)
        self.max_num_contacts = len(self.geom_id_pairs)

    def compute_qp_inequalities(
        self,
        configuration: Configuration,
        dt: float,
    ) -> Constraint:
        upper_bound = np.full((self.max_num_contacts,), np.inf)
        coefficient_matrix = np.zeros((self.max_num_contacts, self.model.nv))
        for idx, (geom1_id, geom2_id) in enumerate(self.geom_id_pairs):
            contact = self._compute_contact_with_minimum_distance(
                configuration.data, geom1_id, geom2_id
            )
            if contact.inactive:
                continue
            hi_bound_dist = contact.dist
            if hi_bound_dist > self.minimum_distance_from_collisions:
                dist = hi_bound_dist - self.minimum_distance_from_collisions
                upper_bound[idx] = (self.gain * dist / dt) + self.bound_relaxation
            else:
                upper_bound[idx] = self.bound_relaxation
            jac = compute_contact_normal_jacobian(
                self.model, configuration.data, contact
            )
            coefficient_matrix[idx] = -jac
        return Constraint(G=coefficient_matrix, h=upper_bound)

    # Private methods.

    def _compute_contact_with_minimum_distance(
        self, data: mujoco.MjData, geom1_id: int, geom2_id: int
    ) -> Contact:
        """Returns the smallest signed distance between a geom pair."""
        fromto = np.empty(6)
        dist = mujoco.mj_geomDistance(
            self.model,
            data,
            geom1_id,
            geom2_id,
            self.collision_detection_distance,
            fromto,
        )
        return Contact(
            dist, fromto, geom1_id, geom2_id, self.collision_detection_distance
        )

    def _homogenize_geom_id_list(self, geom_list: GeomSequence) -> List[int]:
        """Take a heterogeneous list of geoms (specified via ID or name) and return
        a homogenous list of IDs (int)."""
        list_of_int: list[int] = []
        for g in geom_list:
            if isinstance(g, int):
                list_of_int.append(g)
            else:
                assert isinstance(g, str)
                list_of_int.append(self.model.geom(g).id)
        return list_of_int

    def _collision_pairs_to_geom_id_pairs(self, collision_pairs: CollisionPairs):
        geom_id_pairs = []
        for collision_pair in collision_pairs:
            id_pair_A = self._homogenize_geom_id_list(collision_pair[0])
            id_pair_B = self._homogenize_geom_id_list(collision_pair[1])
            id_pair_A = list(set(id_pair_A))
            id_pair_B = list(set(id_pair_B))
            geom_id_pairs.append((id_pair_A, id_pair_B))
        return geom_id_pairs

    def _construct_geom_id_pairs(self, geom_pairs):
        """Returns a set of geom ID pairs for all possible geom-geom collisions.

        The contacts are added based on the following heuristics:
            1) Geoms that are part of the same body or weld are not included.
            2) Geoms where the body of one geom is a parent of the body of the other
                geom are not included.
            3) Geoms that fail the contype-conaffinity check are ignored.

        Note:
            1) If two bodies are kinematically welded together (no joints between them)
                they are considered to be the same body within this function.
        """
        geom_id_pairs = []
        for id_pair in self._collision_pairs_to_geom_id_pairs(geom_pairs):
            for geom_a, geom_b in itertools.product(*id_pair):
                weld_body_cond = not _is_welded_together(self.model, geom_a, geom_b)
                parent_child_cond = not _are_geom_bodies_parent_child(
                    self.model, geom_a, geom_b
                )
                contype_conaffinity_cond = _is_pass_contype_conaffinity_check(
                    self.model, geom_a, geom_b
                )
                if weld_body_cond and parent_child_cond and contype_conaffinity_cond:
                    geom_id_pairs.append((min(geom_a, geom_b), max(geom_a, geom_b)))
        return geom_id_pairs



"""Center-of-mass task implementation."""

from __future__ import annotations

from typing import Optional

import mujoco
import numpy as np
import numpy.typing as npt

from ..configuration import Configuration
from .exceptions import InvalidTarget, TargetNotSet, TaskDefinitionError
from .task import Task


class ComTask(Task):
    """Regulate the center-of-mass (CoM) of a robot.

    Attributes:
        target_com: Target position of the CoM.
    """

    k: int = 3
    target_com: Optional[np.ndarray]

    def __init__(
        self,
        cost: npt.ArrayLike,
        gain: float = 1.0,
        lm_damping: float = 0.0,
    ):
        super().__init__(cost=np.zeros((self.k,)), gain=gain, lm_damping=lm_damping)
        self.target_com = None

        self.set_cost(cost)

    def set_cost(self, cost: npt.ArrayLike) -> None:
        cost = np.atleast_1d(cost)
        if cost.ndim != 1 or cost.shape[0] not in (1, self.k):
            raise TaskDefinitionError(
                f"{self.__class__.__name__} cost must be a vector of shape (1,) "
                f"(aka identical cost for all coordinates) or ({self.k},). "
                f"Got {cost.shape}"
            )
        if not np.all(cost >= 0.0):
            raise TaskDefinitionError(f"{self.__class__.__name__} cost must be >= 0")
        self.cost[:] = cost

    def set_target(self, target_com: npt.ArrayLike) -> None:
        """Set the target CoM position in the world frame.

        Args:
            target_com: Desired center-of-mass position in the world frame.
        """
        target_com = np.atleast_1d(target_com)
        if target_com.ndim != 1 or target_com.shape[0] != (self.k):
            raise InvalidTarget(
                f"Expected target CoM to have shape ({self.k},) but got "
                f"{target_com.shape}"
            )
        self.target_com = target_com.copy()

    def set_target_from_configuration(self, configuration: Configuration) -> None:
        """Set the target CoM from a given robot configuration.

        Args:
            configuration: Robot configuration :math:`q`.
        """
        self.set_target(configuration.data.subtree_com[1])

    def compute_error(self, configuration: Configuration) -> np.ndarray:
        """Compute the CoM task error.

        Args:
            configuration: Robot configuration :math:`q`.

        Returns:
            Center-of-mass task error vector :math:`e(q)`.
        """
        if self.target_com is None:
            raise TargetNotSet(self.__class__.__name__)
        return configuration.data.subtree_com[1] - self.target_com

    def compute_jacobian(self, configuration: Configuration) -> np.ndarray:
        """Compute the CoM task Jacobian.

        Args:
            configuration: Robot configuration :math:`q`.

        Returns:
            Center-of-mass task jacobian :math:`J(q)`.
        """
        if self.target_com is None:
            raise TargetNotSet(self.__class__.__name__)
        jac = np.empty((self.k, configuration.nv))
        mujoco.mj_jacSubtreeCom(configuration.model, configuration.data, jac, 1)
        return jac



from setuptools import setup, find_packages
import os


setup(
    name='dart_physics',
    version='1.6.8',
    packages=find_packages(),
    install_requires=[
        'mujoco',
        'gdown',
        'adam-robotics[jax]==0.3.0',
        'dm_control',
        'loop_rate_limiters',
        'avp_stream',
        'robot_descriptions',
        'obj2mjcf',
        'flask>=3.0.3',
        'psutil', 
        'mediapy', 
        'pyzmq', 
        'grpcio',
        'grpcio-tools',
        'numpy',
        'imageio>=2.36.0',
        'opencv-python',
        'qpsolvers[quadprog] >= 4.3.1',
        'typing_extensions',
        'dexhub-api>=0.3',
        'gdown', 
    ],
    author='Younghyo Park',
    author_email='younghyo@mit.edu',
    python_requires='>=3.6',
)




from dart_physics.robot_cfgs import * 
from pathlib import Path

_HERE = Path(__file__).resolve().parent


def qpos2ctrl(model, data, qpos_target): 

    ctrl = np.zeros(25)
    ctrl = qpos_target[:25]
    return ctrl 


bodies = ["left_shoulder_pitch_link", \
          "left_shoulder_roll_link", "left_shoulder_yaw_link", \
            "left_elbow_pitch_link", "left_elbow_roll_link", \
            "left_zero_link", "left_one_link", "left_two_link", \
            "left_three_link", "left_four_link", "left_five_link", \
            "left_six_link"] \
        + ["right_shoulder_pitch_link", "right_shoulder_roll_link", \
            "right_shoulder_yaw_link", "right_elbow_pitch_link", \
            "right_elbow_roll_link", "right_zero_link", "right_one_link", \
            "right_two_link", "right_three_link", "right_four_link", \
            "right_five_link", "right_six_link"]

robot_cfg = {

    "name": "DualPanda",

    "robots": { 
        "robot": RobotConfig( 
                    xml_path = (_HERE / "../assets" / "unitree_g1" / "g1.xml").as_posix(),
                    home_q = [0.0] * 25, 
                    freejoint = False, 
                    attach_to = [0.0, 0.0, 0.2, 0.7071068 , 0, 0, 0.7071068 ], 
                    parallel_jaw = False, 
                    add_bodies = [ 
                        NewBodyConfig(attach_to = "right_six_link", \
                                      body_name = "right_six_link_tip", pos = [0.05, 0.0, 0.0]),
                        NewBodyConfig(attach_to = "right_four_link", \
                                        body_name = "right_four_link_tip", pos = [0.05, 0.0, 0.0]), 
                        NewBodyConfig(attach_to = "right_two_link", \
                                        body_name = "right_two_link_tip", pos = [0.0, 0.05, 0.0]),
                        NewBodyConfig(attach_to = "right_zero_link", \
                                        body_name = "right_zero_link_tip", pos = [-0.05, 0, 0]),

                        NewBodyConfig(attach_to = "left_six_link", \
                                        body_name = "left_six_link_tip", pos = [0.05, 0.0, 0.0]),
                        NewBodyConfig(attach_to = "left_four_link", \
                                        body_name = "left_four_link_tip", pos = [0.05, 0.0, 0.0]),
                        NewBodyConfig(attach_to = "left_two_link", \
                                        body_name = "left_two_link_tip", pos = [0.0, -0.05, 0.0]),
                        NewBodyConfig(attach_to = "left_zero_link", \
                                        body_name = "left_zero_link_tip", pos = [-0.05, 0, 0]),
                    ], 
                    
                ), 
    }, 

    "qpos2ctrl": qpos2ctrl, 
    "obj_startidx": (7+2) * 2, 
        
    "points" : { 

        "left_wrist":     FingerKeypoints("left", 0, "left_wrist_target", "robot/left_zero_link_tip", type = "body", avp_transform = rot_x(90)),
        "left_two":      FingerKeypoints("left", 4, "left_two_target", "robot/left_two_link_tip", type = "body"),
        "left_four":      FingerKeypoints("left", 9, "left_four_target", "robot/left_four_link_tip", type = "body"),
        "left_six":      FingerKeypoints("left", 14, "left_six_target", "robot/left_six_link_tip", type = "body"),

        "right_wrist":     FingerKeypoints("right", 0, "right_wrist_target", "robot/right_zero_link_tip", type = "body", avp_transform = rot_x(90) @ rot_y(180)),
        "right_two":      FingerKeypoints("right", 4, "right_two_target", "robot/right_two_link_tip", type = "body"),
        "right_four":      FingerKeypoints("right", 9, "right_four_target", "robot/right_four_link_tip", type = "body"),
        "right_six":      FingerKeypoints("right", 14, "right_six_target", "robot/right_six_link_tip", type = "body"),

    },

    "avp_calib": { 

        "left_wrist": {"scale": 1.0, "offset": np.array([0.0, 0.0, 0.0])},
        "left_two": {"scale": 1.0, "offset": np.array([0.0, 0.0, 0.0])},
        "left_four": {"scale": 1.0, "offset": np.array([0.0, 0.0, 0.0])},
        "left_six": {"scale": 1.0, "offset": np.array([0.0, 0.0, 0.0])},

        "right_wrist": {"scale": 1.0, "offset": np.array([0.0, 0.0, 0.0])},
        "right_two": {"scale": 1.0, "offset": np.array([0.0, 0.0, 0.0])},
        "right_four": {"scale": 1.0, "offset": np.array([0.0, 0.0, 0.0])},
        "right_six": {"scale": 1.0, "offset": np.array([0.0, 0.0, 0.0])},
    }, 

    "ik_task" : [
        # IKTasks(target = "left_wrist", pos_cost = 1.0, ori_cost = 0.5), 
        # IKTasks(root = "left_wrist", target = "left_two"),
        # IKTasks(root = "left_wrist", target = "left_four"),
        # IKTasks(root = "left_wrist", target = "left_six"),

        # # IKTasks(target = "right_wrist", pos_cost = 1.0, ori_cost = 0.5),
        # IKTasks(root = "right_wrist", target = "right_two"),
        # IKTasks(root = "right_wrist", target = "right_four"),
        # IKTasks(root = "right_wrist", target = "right_six"),

        IKTasks(target = "left_wrist"), 
        IKTasks(target = "left_two"), 
        IKTasks(target = "left_four"), 
        IKTasks(target = "left_six"),

        IKTasks(target = "right_wrist"),
        IKTasks(target = "right_two"), 
        IKTasks(target = "right_four"), 
        IKTasks(target = "right_six"), 
    ],

    "joints" : ["robot/torso_joint",
                            
                "robot/left_shoulder_pitch_joint",
                "robot/left_shoulder_roll_joint",
                "robot/left_shoulder_yaw_joint",
                "robot/left_elbow_pitch_joint",
                "robot/left_elbow_roll_joint",

                "robot/right_shoulder_pitch_joint",
                "robot/right_shoulder_roll_joint",
                "robot/right_shoulder_yaw_joint",
                "robot/right_elbow_pitch_joint",
                "robot/right_elbow_roll_joint",

                ], 

    "bodies":  {f"robot/{key}":f"gggg1_{key}" for key in bodies}, 

    "usdz_geoms":  [ 
        

        # LEFT 

        BodyInfo(
            name = "g1_left_shoulder_pitch_link",
            geoms = [GeomInfo(mesh = "left_shoulder_pitch_link")],
            pos="-0.0025 0.10396 0.25928", quat="0.990268 0.139172 0 0"
        ), 

        BodyInfo(
            name = "g1_left_shoulder_roll_link",
            geoms = [GeomInfo(mesh = "left_shoulder_roll_link")],
            pos="0 0.052 0", quat="0.990268 -0.139172 0 0", 
        ), 

        BodyInfo(
            name = "g1_left_shoulder_yaw_link",
            geoms = [GeomInfo(mesh = "left_shoulder_yaw_link")],
            pos="-0.00354 0.0062424 -0.1032", 
        ), 

        BodyInfo(
            name = "g1_left_elbow_pitch_link",
            geoms = [GeomInfo(mesh = "left_elbow_pitch_link")],
            pos="0 0.00189 -0.0855", 
        ), 

        BodyInfo(
            name = "g1_left_elbow_roll_link",
            geoms = [GeomInfo(mesh = "left_elbow_roll_link"), GeomInfo(mesh = "left_palm_link", pos=[0.12, 0, 0])],
            pos="0.1 0 0", 
        ),


        BodyInfo(
            name = "g1_left_zero_link",
            geoms = [GeomInfo(mesh = "left_zero_link")],
            pos="0.17 0 0", 
        ), 

        BodyInfo(
            name = "g1_left_one_link",
            geoms = [GeomInfo(mesh = "left_one_link")],
            pos="-0.026525 -0.0188 -5e-05", 
        ), 

        BodyInfo(
            name = "g1_left_two_link",
            geoms = [GeomInfo(mesh = "left_two_link")],
            pos="0 -0.0431 0", 
        ), 

        BodyInfo(
            name = "g1_left_three_link",
            geoms = [GeomInfo(mesh = "left_three_link")],
            pos="0.205 0.004 0.02395"
        ), 

        BodyInfo(
            name = "g1_left_four_link",
            geoms = [GeomInfo(mesh = "left_four_link")],
            pos="0.0471 -0.0036 0", 
        ), 

        BodyInfo(
            name = "g1_left_five_link", 
            geoms = [GeomInfo(mesh = "left_five_link")],
            pos="0.205 0.004 -0.02395", 
        ), 

        BodyInfo(
            name = "g1_left_six_link",
            geoms = [GeomInfo(mesh = "left_six_link")],
            pos="0.0471 -0.0036 0", 
        ), 

        # RIGHT 

        BodyInfo(
            name = "g1_right_shoulder_pitch_link",
            geoms = [GeomInfo(mesh = "right_shoulder_pitch_link")],
            pos="-0.0025 -0.10396 0.25928", quat="0.990268 -0.139172 0 0"
        ), 



        BodyInfo(
            name = "g1_right_shoulder_roll_link",
            geoms = [GeomInfo(mesh = "right_shoulder_roll_link")],
            pos="0 -0.052 0", quat="0.990268 0.139172 0 0"
        ), 

        BodyInfo(
            name = "g1_right_shoulder_yaw_link",
            geoms = [GeomInfo(mesh = "right_shoulder_yaw_link")],
            pos="-0.00354 -0.0062424 -0.1032",
        ), 

        BodyInfo(
            name = "g1_right_elbow_pitch_link",
            geoms = [GeomInfo(mesh = "right_elbow_pitch_link")],
            pos="0 -0.00189 -0.0855",
        ), 

        BodyInfo(
            name = "g1_right_elbow_roll_link",
            geoms = [GeomInfo(mesh = "right_elbow_roll_link"), GeomInfo(mesh = "right_palm_link", pos=[0.12, 0, 0])],
            pos="0.1 0 0",
        ),

        BodyInfo(
            name = "g1_right_zero_link",
            geoms = [GeomInfo(mesh = "right_zero_link")],
            pos="0.17 0 0",
        ),

        BodyInfo(
            name = "g1_right_one_link",
            geoms = [GeomInfo(mesh = "right_one_link")],
            pos="-0.026525 0.0188 -5e-05",
        ),

        BodyInfo(
            name = "g1_right_two_link",
            geoms = [GeomInfo(mesh = "right_two_link")],
            pos="0 0.0431 0",
        ),

        BodyInfo(
            name = "g1_right_three_link",
            geoms = [GeomInfo(mesh = "right_three_link")],
            pos="0.205 -0.004 0.02395"
        ),

        BodyInfo(
            name = "g1_right_four_link",
            geoms = [GeomInfo(mesh = "right_four_link")],
            pos="0.0471 0.0036 0",
        ),

        BodyInfo(
            name = "g1_right_five_link",
            geoms = [GeomInfo(mesh = "right_five_link")],
            pos="0.205 -0.004 -0.02395",
        ),

        BodyInfo(
            name = "g1_right_six_link",
            geoms = [GeomInfo(mesh = "right_six_link")],
            pos="0.0471 0.0036 0",
        ),

    ]



} 


if __name__ == "__main__": 

    import trimesh 
    from scipy.spatial.transform import Rotation as R
    import os 

    asset_root = os.path.join(_HERE, "../assets", "unitree_g1", "assets")
    usdz_root = os.path.join(_HERE, "../assets", "unitree_g1", "transformed_stl3")\

    os.makedirs(usdz_root, exist_ok = True)

    input_list = [f"{k}" for k in robot_cfg["bodies"].values()]
    output_string = str(input_list).replace("'", '"')
    print(output_string)

    for body in robot_cfg["usdz_geoms"]:

        body: BodyInfo

        meshes = [] 
        for i, geom in enumerate(body.geoms): 
            geom: GeomInfo
            mesh = trimesh.load_mesh(f"{asset_root}/{geom.mesh}.stl")
            pos = np.array(geom.pos)
            quat = np.array(geom.quat)

            mat = np.eye(4)
            mat[:3, :3] = R.from_quat(quat, scalar_first = True).as_matrix()
            mat[:3, 3] = pos

            mesh.apply_transform(mat)

            meshes.append(mesh)
        
        meshes = trimesh.util.concatenate(meshes)
        # mat = np.eye(4)
        # pos = np.array(body.pos)
        # quat = np.array(body.quat)

        # mat[:3, :3] = R.from_quat(quat, scalar_first = True).as_matrix()
        # mat[:3, 3] = pos

        # meshes.apply_transform(mat)

        meshes.export(f"{usdz_root}/{body.name}.obj")





import numpy as np


def get_epsilon(dtype: np.dtype) -> float:
    return {
        np.dtype("float32"): 1e-5,
        np.dtype("float64"): 1e-10,
    }[dtype]


def skew(x: np.ndarray) -> np.ndarray:
    assert x.shape == (3,)
    wx, wy, wz = x
    return np.array(
        [
            [0.0, -wz, wy],
            [wz, 0.0, -wx],
            [-wy, wx, 0.0],
        ],
        dtype=x.dtype,
    )



"""Exceptions raised by tasks."""

from ..exceptions import MinkError


class TaskDefinitionError(MinkError):
    """Exception raised when a task definition is ill-formed."""


class TargetNotSet(MinkError):
    """Exception raised when attempting to use a task with an unset target."""

    def __init__(self, cls_name: str):
        message = f"No target set for {cls_name}"
        super().__init__(message)


class InvalidTarget(MinkError):
    """Exception raised when the target is invalid."""


class InvalidGain(MinkError):
    """Exception raised when the gain is outside the valid range."""


class InvalidDamping(MinkError):
    """Exception raised when the damping is outside the valid range."""



import platform
from flask import Flask, request, jsonify
import subprocess
import signal
import os 
import psutil
import yaml
import logging
from logging.handlers import RotatingFileHandler
from datetime import datetime
import threading
import base64
import os
from dart_physics.runs import load_task_cfg, load_robot_cfg
from pathlib import Path
import zipfile
import gdown 

CUR_PATH = Path(__file__).resolve().parent

ASSET_URL = "https://drive.google.com/file/d/1U9YeVJfabrSyXPmVQzrxIscDciik3rp4/view?usp=sharing"


def download_and_extract_assets():
    # Import gdown after it's installed
    # Resolve the current directory to find where dart_physics is installed
    CUR_PATH = Path(__file__).resolve().parent
    ASSET_DIR = os.path.join(CUR_PATH)
    ZIP_FILE_PATH = os.path.join(ASSET_DIR, "dart_physics_assets.zip")

    if not os.path.exists(ASSET_DIR):
        os.makedirs(ASSET_DIR)
    
    # Check if assets are already downloaded
    if not os.path.exists(ZIP_FILE_PATH):
        print(f"Downloading assets from {ASSET_URL}...")
        gdown.download(ASSET_URL, ZIP_FILE_PATH, quiet=False, fuzzy=True)
        print("Assets downloaded.")
    else:
        print("Assets already exist, skipping download.")

    # Unzip the downloaded file
    if not os.path.exists(os.path.join(ASSET_DIR, 'unzipped_assets_folder')):  # Check if already unzipped
        print("Unzipping assets...")
        with zipfile.ZipFile(ZIP_FILE_PATH, 'r') as zip_ref:
            zip_ref.extractall(ASSET_DIR)
        print("Assets unzipped.")
    else:
        print("Assets already unzipped.")

def get_os():
    os_name = platform.system()
    if os_name == "Darwin":
        return "macOS"
    elif os_name == "Linux":
        return "Linux (Ubuntu or other)"
    else:
        return f"Other: {os_name}"


download_and_extract_assets()

IS_MAC = get_os() == "macOS"
print("USING MAC?", IS_MAC)

app = Flask(__name__)
current_process = None

# Set up logging
log_file = 'subprocess_output.log'
logger = logging.getLogger('SubprocessLogger')
logger.setLevel(logging.INFO)
handler = RotatingFileHandler(log_file, maxBytes=10*1024*1024, backupCount=5)  # Rotate after 10 MB, keep 5 backups
formatter = logging.Formatter('%(asctime)s - %(message)s')
handler.setFormatter(formatter)
logger.addHandler(handler)

robot_script_mapper = { 
    "Single FR3": "left_panda",
    "Single Panda": "left_panda", 
    "Dual Panda": "dual_panda",
    "ALOHA": "aloha",
    "Unitree G1": "unitree_g1", 
    "Rainbow Y1": "rby1",
    "Allegro Hand": "freefloating_dual_allegro"
}

@app.route('/run-script', methods=['POST'])
def run_script():
    global current_process
    try:
        # Get the client's IP address

        # Get the arguments from the request JSON
        data = request.get_json()
        task = data.get('task')
        robot = data.get('robot', 'Dual Panda')
        extra = data.get('extra', '')
        mobile = data.get('mobile', "Fixed Base")

        if mobile == "Mobile": 
            robot_file_name = robot_script_mapper[robot] + "_mobile"
        else: 
            robot_file_name = robot_script_mapper[robot]

        client_ip = data.get('ip')

        logger.info(f'Request received from IP: {client_ip}')
        
        if not task:
            return jsonify({'error': 'task is required'}), 400


        my_env  = os.environ.copy()


        # Construct the command with the provided task and client IP
        if IS_MAC:
            command = ['mjpython', '-m', 'dart_physics.runs.teleop', '--task', task, '--robot', robot_file_name, '--avp', '--th', extra] 
        else:
            command = ['python', '-m','dart_physics.runs.teleop', '--task', task, '--robot', robot_file_name, '--headless', '--avp', '--th', extra] 


        # Run the script using subprocess
        logger.info(f"command: {command}") 
        current_process = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True, env = my_env) 
        logger.info(f'started MuJoCo process.') 

        def log_subprocess_output(pipe):
            for line in iter(pipe.readline, ''):
                logger.info(line.strip())

        # Log stdout and stderr asynchronously 
        stdout_thread = threading.Thread(target=log_subprocess_output, args=(current_process.stdout,))
        stderr_thread = threading.Thread(target=log_subprocess_output, args=(current_process.stderr,))
        stdout_thread.start()
        stderr_thread.start()

        # Wait for the process to complete 
        # current_process.wait()
        # stdout_thread.join()
        # stderr_thread.join()

        return jsonify({'message': 'Script started successfully', 'client_ip': client_ip})
    
    except Exception as e:
        logger.error(f'Error running script: {str(e)}')
        return jsonify({'error': str(e)}), 500


@app.route('/kill-script', methods=['POST'])
def kill_script():
    try:
        
        # Iterate through all running processes
        for proc in psutil.process_iter(['pid', 'name', 'cmdline']):
            try:
                # Check if the process command line matches 'python run.py'
                cmdline = proc.info['cmdline']
                if cmdline and cmdline[0] == 'python' and 'runs/teleop.py' in cmdline:
                    # Terminate the process
                    proc.terminate()
                    proc.wait()
                    logger.info(f'Terminated process {proc.pid}')
            except (psutil.NoSuchProcess, psutil.AccessDenied, psutil.ZombieProcess):
                pass

        return jsonify({'message': 'All run.py scripts terminated successfully'})
    
    except Exception as e:
        logger.error(f'Error terminating script: {str(e)}')
        return jsonify({'error': str(e)}), 500


@app.route('/list-files', methods=['GET'])
def list_files():
    logger.info(f'Got request to list files.')
    cfg_path = CUR_PATH / "cfgs"
    try:
        files = os.listdir(cfg_path)
        yaml_files = [f[:-5] for f in files if f.endswith('.yaml')] +  [f[:-3] for f in files if f.endswith('.py') and f != "__init__.py"]

        # if dummy.py exists, remove it from the list 
        if 'dummy' in yaml_files:
            yaml_files.remove('dummy')

        # List PNG files
        images = {}
        image_files = [f[:-4] for f in files if f.endswith('.png')]
        for img_file in image_files:
            with open(os.path.join(cfg_path, img_file+'.png'), "rb") as image_file:
                encoded_string = base64.b64encode(image_file.read()).decode('utf-8')
                images[img_file] = encoded_string
        logger.info(f'Files listed: {files} and images: {list(images.keys())}')
        return jsonify({'files': yaml_files, 'images': images})
    
    except Exception as e:
        logger.error(f'Error listing files: {str(e)}')
        return jsonify({'error': str(e)}), 500

@app.route('/get-objects', methods=['POST'])
def get_objects():
    """
    # request input:  {"task": "task_name"}
    # response output: {"objects": ["object1", "object2", ...]}
    # retreive object by reading the yaml file 
    # yaml file located in "../sim/cfgs/task_name.yaml"
    # look for key "objects" in the yaml file
    """

    try:
        data = request.get_json()
        task = data.get('task')
        if not task:
            return jsonify({'error': 'task is required'}), 400

        
        # file_path = f"cfgs/{task}.yaml"
        # with open(file_path, 'r') as file:
        #     cfg = yaml.safe_load(file)

        cfg = load_task_cfg(task)

        objects = cfg.get('objects', [])

        objects_to_load = [] 
        for obj in objects: 

            body_dict = cfg.get('object_meshes',  {})
            if obj in body_dict.keys():
                bodies = body_dict[obj]
                
                for body in bodies: 
                    objects_to_load.append(obj + '_' + body)
            else:
                objects_to_load.append(obj)

        scales = cfg.get('object_scale', {obj: 1.0 for obj in objects_to_load})
        logger.info(f'Objects for task {task}: {objects_to_load}')

        return jsonify({'objects': objects_to_load, 'scales': scales})
    
    except Exception as e:
        logger.error(f'Error getting objects: {str(e)}')
        return jsonify({'error': str(e)}), 500

def handle_shutdown(signal, frame):
    global current_process
    if current_process:
        current_process.terminate()
        current_process.wait()
    os._exit(0)


@app.route('/status', methods=['GET'])
def get_status():
    logger.info(f'Got request to check status.')
    try:
        return jsonify({'status': 'Active'})
    except Exception as e:
        logger.error(f'Error getting status: {str(e)}')
        return jsonify({'error': str(e)}), 500

if __name__ == '__main__':
    # Register the signal handlers
    signal.signal(signal.SIGINT, handle_shutdown)
    signal.signal(signal.SIGTERM, handle_shutdown)
    
    app.run(host='0.0.0.0', port=5012)




from dart_physics.robot_cfgs import * 
from pathlib import Path
import mujoco
import dart_physics.mink as mink 

_HERE = Path(__file__).resolve().parent


def qpos2ctrl(model, data, qpos_target): 

    ctrl = np.zeros(8)
    ctrl[:7] = qpos_target[0:7]

    # gripper control
    tip1 = mink.SE3.from_mocap_name(model, data, "right_lf_target").as_matrix()
    tip2 = mink.SE3.from_mocap_name(model, data, "right_rf_target").as_matrix()

    finger_dist = np.linalg.norm(tip1[:3, 3] - tip2[:3, 3])

    max_grip = data.qpos[8] - 0.01

    ctrl[-1] = np.clip(finger_dist - 0.01, max_grip, 0.04)

    return ctrl 


robot_cfg = {

    "name": "right_panda",

    "robots": { 
        "r_robot": RobotConfig( 
                    xml_path = (_HERE / "../assets" / "franka_emika_panda" / "new_panda.xml").as_posix(),
                    home_q = [0, 0, 0, -1.5707899999999999, 0, 1.5707899999999999, -0.7853, 0.040000000000000001, 0.040000000000000001], 
                    freejoint = False, 
                    attach_to = [+0.4, -0.15, 0.0, 0.7071068 , 0, 0, 0.7071068 ], 
                    parallel_jaw = True, 
                    add_bodies = [ 
                        NewBodyConfig(attach_to = "right_finger", \
                                      body_name = "right_finger_base", pos = [0.0, 0.02, 0.0]),
                        NewBodyConfig(attach_to = "right_finger", \
                                        body_name = "right_finger_tip", pos = [0, 0, 0.05]), 
                        NewBodyConfig(attach_to = "left_finger", \
                                        body_name = "left_finger_base", pos = [0.0, 0.02, 0.0]),
                        NewBodyConfig(attach_to = "left_finger", \
                                        body_name = "left_finger_tip", pos = [0, 0, 0.05]),
                    ], 
                    
                ), 
    }, 

    "qpos2ctrl": qpos2ctrl, 
    "obj_startidx": (7+2) , 
        
    "points" : { 

        "right_lb":     FingerKeypoints("right", 3, "right_lb_target", "r_robot/left_finger_base", type = "body"),
        "right_lf":     FingerKeypoints("right", 4, "right_lf_target", "r_robot/left_finger_tip", type = "body"),
        "right_rb":     FingerKeypoints("right", 8, "right_rb_target", "r_robot/right_finger_base", type = "body"),
        "right_rf":     FingerKeypoints("right", 9, "right_rf_target", "r_robot/right_finger_tip", type = "body"),
    },

    "avp_calib": { 

        "right_lb": {"scale": 1.0, "offset": np.array([0.0, 0.0, 0.0])},
        "right_lf": {"scale": 1.0, "offset": np.array([0.0, 0.0, 0.0])},
        "right_rb": {"scale": 1.0, "offset": np.array([0.0, 0.0, 0.0])},
        "right_rf": {"scale": 1.0, "offset": np.array([0.0, 0.0, 0.0])},

    }, 

    "ik_task" : [
        IKTasks(target = "right_lb"), 
        IKTasks(target = "right_lf", pos_cost = 1.5),
        IKTasks(target = "right_rb"),
        IKTasks(target = "right_rf", pos_cost = 1.5),
    ],

    "joints" : ['r_robot/joint{}'.format(i) for i in range(1, 8)], 

    "bodies":  {
        # right robot
        "r_robot/link0":   "panda_link0_right",   "r_robot/link1":   "panda_link1_right",
        "r_robot/link2":   "panda_link2_right",   "r_robot/link3":   "panda_link3_right",
        "r_robot/link4":   "panda_link4_right",   "r_robot/link5":   "panda_link5_right",
        "r_robot/link6":   "panda_link6_right",   "r_robot/link7":   "panda_link7_right",
        "r_robot/hand":    "panda_hand_right",    "r_robot/left_finger": "panda_left_finger_right",
        "r_robot/right_finger": "panda_right_finger_right",
    } 

} 
